# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import json
from pathlib import Path
import numpy as np
import pandas as pd

from .beam_elements import BaseCollimator, BlackAbsorber, EverestCollimator, EverestCrystal, _all_collimator_types, element_classes
from .colldb import CollimatorDatabase
from .impacts import CollimatorImpacts
from .scattering_routines.everest.materials import SixTrack_to_xcoll, CrystalMaterial

import xobjects as xo
import xpart as xp
import xtrack as xt


# Logic of buffers:
#    A buffer is a continuous stream of memory, containing a bunch of xobjects.
#    It is an architecture-invariant way to work with pointers, as the pointers to the individual
#    xobjects are stored as offsets to the start of the buffer.
#    The tracker manages three buffers: one for the line and all elements, one for the particles,
#    and an io_buffer for record logging.

class CollimatorManager:

    _init_vars = ['_colldb', 'line', 'beam', 'capacity', 'record_impacts', '_context', '_buffer', 'io_buffer', '_part_buffer']
    _init_var_defaults = {'_colldb': None, 'beam': None, 'capacity': 1e6, 'record_impacts': False, \
                          '_context': None, '_buffer': None, 'io_buffer': None, '_part_buffer': None}

    # -------------------------------
    # ------ Loading functions ------
    # -------------------------------

    @classmethod
    def from_yaml(cls, file, **kwargs):
        if '_colldb' in kwargs.keys():
            raise ValueError("Cannot load CollimatorDatabase file and specify '_colldb' argument in loader!")
        kwargs_colldb  = {key: val for key, val in kwargs.items() if key in CollimatorDatabase._init_vars}
        kwargs_manager = {key: val for key, val in kwargs.items() if key in cls._init_vars}
        colldb = CollimatorDatabase.from_yaml(file, **kwargs_colldb)
        return cls(_colldb=colldb, **kwargs_manager)

    @classmethod
    def from_json(cls, file, **kwargs):
        if '_colldb' in kwargs.keys():
            raise ValueError("Cannot load CollimatorDatabase file and specify '_colldb' argument in loader!")
        kwargs_colldb  = {key: val for key, val in kwargs.items() if key in CollimatorDatabase._init_vars}
        kwargs_manager = {key: val for key, val in kwargs.items() if key in cls._init_vars}
        colldb = CollimatorDatabase.from_json(file, **kwargs_colldb)
        return cls(_colldb=colldb, **kwargs_manager)

    @classmethod
    def from_dict(cls, file, **kwargs):
        if '_colldb' in kwargs.keys():
            raise ValueError("Cannot load CollimatorDatabase file and specify '_colldb' argument in loader!")
        kwargs_colldb  = {key: val for key, val in kwargs.items() if key in CollimatorDatabase._init_vars}
        kwargs_manager = {key: val for key, val in kwargs.items() if key in cls._init_vars}
        colldb = CollimatorDatabase.from_dict(file, **kwargs_colldb)
        return cls(_colldb=colldb, **kwargs_manager)

    @classmethod
    def from_SixTrack(cls, file, **kwargs):
        if '_colldb' in kwargs.keys():
            raise ValueError("Cannot load CollimatorDatabase file and specify '_colldb' argument in loader!")
        kwargs_colldb  = {key: val for key, val in kwargs.items() if key in CollimatorDatabase._init_vars}
        kwargs_manager = {key: val for key, val in kwargs.items() if key in cls._init_vars}
        colldb = CollimatorDatabase.from_SixTrack(file, **kwargs_colldb)
        return cls(_colldb=colldb, **kwargs_manager)


    def __init__(self, **kwargs):
        # Get all arguments
        for var in self._init_vars:
            if var in self._init_var_defaults:
                kwargs.setdefault(var, self._init_var_defaults[var])
            elif var not in kwargs.keys():
                raise ValueError(f"CollimatorManager is missing required argument '{var}'!")

        colldb = kwargs['_colldb']
        if not isinstance(colldb, CollimatorDatabase):
            # TODO allow None as empty database
            raise ValueError("The variable '_colldb' needs to be an xcoll CollimatorDatabase object!")
        else:
            self.colldb = colldb

        line = kwargs['line']
        beam = kwargs['beam']
        if isinstance(beam, str):
            beam = int(beam[-1])
        if not isinstance(line, xt.Line):
            raise ValueError("The variable 'line' needs to be an xtrack Line object!")
        else:
            self.line = line
        self.line._needs_rng = True  # TODO not needed if only BlackAbsorbers
        if beam is not None and beam > 1:
            self._line_is_reversed = True
        else:
            self._line_is_reversed = False
        self._machine_length = self.line.get_length()

        # Create _buffer, _context, and _io_buffer
        _buffer  = kwargs['_buffer']
        _context = kwargs['_context']
        if _buffer is None:
            if _context is None:
                _context = xo.ContextCpu()
            _buffer = _context.new_buffer()
        elif _context is not None and _buffer.context != _context:
            raise ValueError("The provided buffer and context do not match! "
                             + "Make sure the buffer is generated inside the provided context, or alternatively, "
                             + "only pass one of _buffer or _context.")
        self._buffer = _buffer
        self._part_buffer  = kwargs['_part_buffer']
        if self._part_buffer is not None and self._part_buffer.context != _context:
            raise ValueError("The provided particle buffer and context do not match! "
                             + "Make sure the particle buffer is generated inside the provided context.")

        # TODO: currently capacity is only for io_buffer (hence for impacts).
        # Do we need it in the _buffer as well?
        self._capacity = int(kwargs['capacity'])
        io_buffer = kwargs['io_buffer']
        if io_buffer is None:
            io_buffer = xt.new_io_buffer(_context=self._buffer.context, capacity=self.capacity)
        elif self._buffer.context != io_buffer._context:
            raise ValueError("The provided io_buffer lives on a different context than the buffer!")
        self._io_buffer = io_buffer

        # Initialise impacts table
        self._recording_elements = []
        self.record_impacts = kwargs['record_impacts']
        self._impacts = None

        # Variables for lossmap
        self._lossmap = None
        self._summary = None
        self._part    = None


    def __getitem__(self, name):
        return self.colldb[name]


    @property
    def machine_length(self):
        return self._machine_length

    @property
    def impacts(self):
        return self._impacts

    @property
    def record_impacts(self):
        return self._record_impacts

    @record_impacts.setter
    def record_impacts(self, record_impacts):
        if record_impacts is False or record_impacts is None:
            record_stop  = [self.line[name] for name in self._recording_elements]
            record_start = False
        elif record_impacts is True:
            record_stop  = []
            record_start = True
        else:
            if not hasattr(record_impacts, '__iter__') or isinstance(record_impacts, str):
                record_impacts = [record_impacts]
            record_stop  = [self.line[name] for name in set(self._recording_elements) - set(record_impacts)]
            record_start = record_impacts

        # We stop logging these elements
        if record_stop:
            if self.line.tracker is not None:
                self.line.tracker._check_invalidated()
            xt.stop_internal_logging(elements=record_stop)
            # Removed the stopped collimators from list of logged elements
            self._recording_elements = list(set(self._recording_elements) - set(record_stop))

        # We will start logging these elements
        self._record_impacts = record_start
        self._set_record_impacts()

    def _set_record_impacts(self):
        record_impacts = self._record_impacts
        if record_impacts is True:
            record_impacts = self.collimator_names
        if record_impacts is False or record_impacts is None:
            record_impacts = []
        if not hasattr(record_impacts, '__iter__') or isinstance(record_impacts, str):
            record_impacts = [record_impacts]
        if record_impacts and np.all([isinstance(self.line[elem], BaseCollimator) for elem in record_impacts]):
            for name in record_impacts:
                if name not in self.line.element_names:
                    raise ValueError(f"Trying to initialise impact table, but collimator {name} not found in line!")
            elements_to_record = list(set(record_impacts) - set(self._recording_elements))
            elements_to_stop = list(set(self._recording_elements) - set(record_impacts))
            if elements_to_stop:
                # These should have been stopped by the setter function above
                raise ValueError("Some elements are recording but are not supposed to!")
            # Any new elements that need recording but aren't recording yet?
            if elements_to_record:
                elements_to_record = [self.line[name] for name in elements_to_record]
                if self.impacts is None:
                    self._impacts = xt.start_internal_logging(io_buffer=self._io_buffer, capacity=self.capacity, \
                                                              elements=elements_to_record)
                else:
                    xt.start_internal_logging(io_buffer=self._io_buffer, capacity=self.capacity, \
                                              record=self.impacts, elements=elements_to_record)
                self._recording_elements = record_impacts
            self.impacts._coll_ids = {self.line.element_names.index(name): name for name in self._recording_elements}

    @property
    def capacity(self):
        return self._capacity

    @capacity.setter
    def capacity(self, capacity):
        capacity = int(capacity)
        if capacity < self.capacity:
            raise NotImplementedError("Shrinking of capacity not yet implemented!")
        elif capacity == self.capacity:
            return
        else:
            self._io_buffer.grow(capacity-self.capacity)
            if self.impacts is not None:
                # TODO: increase capacity of iobuffer AND of _impacts
                raise NotImplementedError

    @property
    def collimator_names(self):
        # TODO: should only sort whenever collimators are added to colldb
        # or when positions are updated
        db = self.colldb._colldb
#         names = list(db[db.active==True].index)
        names = list(db.index)
        if not np.any([s is None for s in db.s_center]):
            names.sort(key=lambda nn: db.loc[nn, 's_center'])
        return names

    @property
    def s_front(self):
        return self.colldb.s_center - self.colldb.active_length/2 - self.colldb.inactive_front

    @property
    def s_active_front(self):
        return self.colldb.s_center - self.colldb.active_length/2

    @property
    def s_center(self):
        return self.colldb.s_center

    @property
    def s_active_back(self):
        return self.colldb.s_center + self.colldb.active_length/2

    @property
    def s_back(self):
        return self.colldb.s_center + self.colldb.active_length/2 + self.colldb.inactive_back

    def install_black_absorbers(self, names=None, *, verbose=False):
        if names is None:
            names = self.collimator_names
        def install_func(thiscoll, name):
            return BlackAbsorber(
                    inactive_front=thiscoll['inactive_front'],
                    inactive_back=thiscoll['inactive_back'],
                    active_length=thiscoll['active_length'],
                    angle=[thiscoll['angle_L'],thiscoll['angle_R']],
                    active=False,
                    _tracking=False,
                    _buffer=self._buffer
                   )
        self._install_collimators(names, install_func=install_func, verbose=verbose)


    def install_everest_collimators(self, names=None, *, verbose=False):
        if names is None:
            names = self.collimator_names
        names = list(names) # Dataframe does not like to be indexed with a set
        df = self.colldb._colldb.loc[names]
        df_coll = df[[c is None for c in df.crystal]]
        df_cry  = df[[c is not None for c in df.crystal]]
        # Do the installations (start with crystals to avoid recompilation)
        if len(df_cry) > 0:
            def install_func(thiscoll, name):
                material = SixTrack_to_xcoll[thiscoll['material']]
                if len(material) < 2:
                    raise ValueError(f"Could not find crystal material definition from variable {thiscoll['material']}!")
                material = material[1]
                if not isinstance(material, CrystalMaterial):
                    raise ValueError(f"The material {material.name} is not a Crystalmaterial!")
                return EverestCrystal(
                        inactive_front=thiscoll['inactive_front'],
                        inactive_back=thiscoll['inactive_back'],
                        active_length=thiscoll['active_length'],
                        angle=[thiscoll['angle_L'],thiscoll['angle_R']],
                        material=material,
                        active=False,
                        _tracking=False,
                        _buffer=self._buffer
                       )
            self._install_collimators(df_cry.index.values, install_func=install_func, verbose=verbose)
        if len(df_coll) > 0:
            def install_func(thiscoll, name):
                return EverestCollimator(
                        inactive_front=thiscoll['inactive_front'],
                        inactive_back=thiscoll['inactive_back'],
                        active_length=thiscoll['active_length'],
                        angle=[thiscoll['angle_L'],thiscoll['angle_R']],
                        material=SixTrack_to_xcoll[thiscoll['material']][0],
                        active=False,
                        _tracking=False,
                        _buffer=self._buffer
                       )
            self._install_collimators(df_coll.index.values, install_func=install_func, verbose=verbose)


    def _install_collimators(self, names, *, install_func, verbose, support_legacy_elements=False):
        # Check that collimator marker exists in Line and CollimatorDatabase,
        # and that tracker is not yet built
        # TODO: need check that all collimators have aperture before and after
        line = self.line
        if not hasattr(names, '__iter__') or isinstance(names, str):
            names = [names]
        df = self.colldb._colldb
        mask = df.index.isin(names)
        for name in names:
            if name not in line.element_names:
                raise Exception(f"Collimator {name} not found in line!")
            elif name not in self.collimator_names:
                raise Exception(f"Warning: Collimator {name} not found in CollimatorDatabase!...")
        if self.tracker_ready:
            raise Exception("Tracker already built!\nPlease install collimators before building "
                          + "tracker!")

        # Loop over collimators to install
        for name in names:

            # Get s positions
            # This cannot go outside the loop as the indices will change!
            ss = line.get_s_position()

            # Get the settings from the CollimatorDatabase
            thiscoll = df.loc[name]
            idx = line.element_names.index(name)
            # Create the collimator element
            newcoll = install_func(thiscoll, name)
            collimator_class = newcoll.__class__

            # Check that collimator is not installed as different type
            # TODO: automatically replace collimator type and print warning
            if isinstance(line[name], tuple(_all_collimator_types - {collimator_class})):
                raise ValueError(f"Trying to install {name} as {collimator_class.__name__},"
                               + f" but it is already installed as {type(line[name]).__name__}!\n"
                               + f"Please reconstruct the line.")

            # Check that collimator is not installed previously
            elif isinstance(line[name], collimator_class):
                if df.loc[name,'collimator_type'] != collimator_class.__name__:
                    raise Exception(f"Something is wrong: Collimator {name} already installed in "
                                  + f"line as {collimator_class.__name__} element, but registered "
                                  + f"in CollimatorDatabase as {df.loc[name, 'collimator_type']}. "
                                  + f"Please reconstruct the line.")
                if verbose: print(f"Collimator {name} already installed. Skipping...")
                continue

            # TODO: only allow Marker elements, no Drifts!!
            #       How to do this with importing a line for MAD-X or SixTrack...?
            elif not isinstance(line[name], (xt.Marker, xt.Drift)) and not support_legacy_elements:
                raise ValueError(f"Trying to install {name} as {collimator_class.__name__},"
                               + f" but the line element to replace is not an xtrack.Marker "
                               + f"(or xtrack.Drift)!\nPlease check the name, or correct the "
                               + f"element.")

            if verbose: print(f"Installing {name:20} as {collimator_class.__name__}")
            # Update the position and type in the CollimatorDatabase
            df.loc[name,'s_center'] = ss[idx]
            df.loc[name,'collimator_type'] = collimator_class.__name__

            # Find apertures and store them
            # TODO: same with cryotanks for FLUKA
            # TODO: use compound info  ->  need full collimator info from MADX
            # TODO: this is all very hacky....
            aper_before = {}
            aper_after = {}
            if f'{name}_mken' in line.element_names\
            and f'{name}_mkex'in line.element_names:
                # TODO what with transformations? How to shift them in s if different?
                aper_before = {nn.replace('mken', 'upstream'): line[nn].copy()
                               for nn in line.element_names if nn.startswith(f'{name}_mken_aper')}
                aper_after  = {nn.replace('mkex', 'downstream'): line[nn].copy()
                               for nn in line.element_names if nn.startswith(f'{name}_mkex_aper')}
            if len(aper_before) == 0:
                # TODO what with transformations? How to shift them in s from centre to start/end?
                aper_before = {nn.replace('_aper', '_upstream_aper'): line[nn].copy()
                               for nn in line.element_names if nn.startswith(f'{name}_aper')}
            if len(aper_after) == 0:
                aper_after  = {nn.replace('_aper', '_downstream_aper'): line[nn].copy()
                               for nn in line.element_names if nn.startswith(f'{name}_aper')}
            if len(aper_before) == 0 or len(aper_after) == 0:
                print(f"Warning: No aperture found for collimator {name}!")

            # Remove stuff at location of collimator
            l = thiscoll['active_length']
            to_remove = []
            i = idx - 1
            # We remove everything between the beginning and end of the collimator except drifts
            while ss[i] >= ss[idx] - l/2:
                el = line[i]
                nn = line.element_names[i]
                if el.__class__.__name__ == 'Drift':
                    i -= 1
                    continue
                if hasattr(el, 'length') and el.length > 0:
                    raise ValueError(f"Found active element {nn} with length "
                                   + f"{el.length} at location inside collimator!")
                # I don't like this class selection...
                if not el.__class__.__name__ in ['Marker', 'SRotation', \
                                       'YRotation', 'XRotation', 'XYShift'] \
                and not el.__class__.__name__.startswith('Limit'):
                    print(f"Warning: Removed active element {nn} "
                        + f"at location inside collimator!")
                to_remove.append(nn)
                i -= 1
            i = idx + 1
            while ss[i] <= ss[idx] + l/2:
                el = line[i]
                nn = line.element_names[i]
                if el.__class__.__name__ == 'Drift':
                    i += 1
                    continue
                if hasattr(el, 'length') and el.length > 0:
                    raise ValueError(f"Found active element {nn} with length "
                                   + f"{el.length} at location inside collimator!")
                # I don't like this class selection...
                if not el.__class__.__name__ in ['Marker', 'SRotation', \
                                       'YRotation', 'XRotation', 'XYShift'] \
                and not el.__class__.__name__.startswith('Limit'):
                    print(f"Warning: Removed active element {line.element_names[i]} "
                        + f"at location inside collimator!")
                to_remove.append(nn)
                i += 1
            for nn in to_remove:
                # TODO: need to update Compounds
                line.element_names.remove(nn)
                line.element_dict.pop(nn)

            # Do the installation
            s_install = df.loc[name,'s_center'] - thiscoll['active_length']/2 \
                            - thiscoll['inactive_front']
            line.insert_element(element=newcoll, name=name, at_s=s_install)

            # Reinstall apertures
            for aper, el in aper_before.items():
                # TODO: need to update Compounds
                line.insert_element(element=el, name=aper, index=name)
            for aper, el in reversed(aper_after.items()):
                # Reversed because of index+1
                # TODO: need to update Compounds
                line.insert_element(element=el, name=aper,
                                    index=line.element_names.index(name)+1)

        self._set_record_impacts()


    @property
    def installed(self):
        return not any([coll is None for coll in self.colldb.collimator_type])


    def align_collimators_to(self, align):
        if not self.installed:
            raise ValueError("Some collimators have not yet been installed.\n"
                             + "Please install all collimators before aligning the collimators.")
        self.colldb.align_to = align

    @property
    def aligned(self):
        return not np.any(
                    [x is None for x in self.colldb._colldb.s_align_front]
                    + [ x is None for x in self.colldb._colldb.s_align_back]
                )


    def build_tracker(self, **kwargs):
        kwargs.setdefault('_buffer', self._buffer)
        kwargs.setdefault('io_buffer', self._io_buffer)
        if kwargs['_buffer'] != self._buffer:
            raise ValueError("Cannot build tracker with different buffer than the CollimatorManager buffer!")
        if kwargs['io_buffer'] != self._io_buffer:
            raise ValueError("Cannot build tracker with different io_buffer than the CollimatorManager io_buffer!")
        if '_context' in kwargs and kwargs['_context'] != self._buffer.context:
            raise ValueError("Cannot build tracker with different context than the CollimatorManager context!")
        self.line.build_tracker(**kwargs)
        self._set_record_impacts()

    @property
    def tracker_ready(self):
        return self.line is not None and self.line.tracker is not None


    def _compute_optics(self, recompute=False):
        if not self.tracker_ready:
            raise Exception("Please build tracker before computing the optics for the openings!")
        line = self.line

        if recompute or not self.colldb._optics_is_ready:
            # TODO: does this fail on Everest? Can the twiss be calculated at the center of the collimator for everest?
#             pos = { *self.s_active_front, *self.s_center, *self.s_active_back }
            pos = list({ *self.s_active_front, *self.s_active_back })
            tw = line.twiss(at_s=pos)
#             tw = tracker.twiss()
            self.colldb._optics = pd.concat([
                                    self.colldb._optics,
                                    pd.DataFrame({
                                            opt: tw[opt] for opt in self.colldb._optics.columns
#                                     opt: [ np.array(tw[opt])[abs(tw['s']-thispos) < 1e-12][0] for thispos in pos ]
#                                         for opt in self.colldb._optics.columns
                                    }, index=pos)
                                ])
            self.colldb.gamma_rel = line.particle_ref._xobject.gamma0[0]


    # The variable 'gaps' is meant to specify temporary settings that will overrule the CollimatorDatabase.
    # As such, its settings will be applied to the collimator elements in the line, but not
    # written to the CollimatorDatabase. Hence two successive calls to set_openings will not be combined,
    # and only the last call will be applied to the line.
    # The variable 'to_parking' will send all collimators that are not listed in 'gaps' to parking.
    # Similarily, the variable 'full_open' will set all openings of the collimators that are not
    # listed in 'gaps' to 1m.
    def set_openings(self, gaps={}, *, recompute_optics=False, to_parking=False, full_open=False, support_legacy_elements=False):
        if not self.tracker_ready:
            raise Exception("Please build tracker before setting the openings!")
        colldb = self.colldb
        if not self.installed:
            raise ValueError("Some collimators have not yet been installed.\n"
                             + "Please install all collimators before setting the openings.")
        if to_parking and full_open:
            raise ValueError("Cannot send collimators to parking and open them fully at the same time!")
        if not self.aligned:
            self.align_collimators_to('front')

        gaps_OLD = colldb.gap
        names = self.collimator_names
        # Override gap if sending to parking
        if to_parking:
            gaps = { **{ name: None for name in names }, **gaps }
        colldb.gap = gaps

        # Get the optics (to compute the opening)
        self._compute_optics(recompute=recompute_optics)
        if not self.colldb._optics_is_ready:
            raise Exception("Something is wrong: not all optics needed for the jaw openings are calculated!")

        # Configure collimators
        line = self.line
        for name in names:
            # Override openings if opening fully
            if full_open and name not in gaps.keys():
                line[name].active = False
            # Apply settings to element
            elif isinstance(line[name], BaseCollimator):
                line[name].ref_x  = colldb.x[name]
                line[name].ref_y  = colldb.y[name]
                line[name].angle  = colldb.angle[name]
                line[name].jaw_L = colldb._colldb.jaw_LU[name]
                line[name].jaw_R = colldb._colldb.jaw_RU[name]
                # TODO
                line[name].side   = colldb.side[name]
                line[name].active = colldb.active[name]
                if isinstance(line[name], (EverestCollimator, EverestCrystal)) or support_legacy_elements:
                    if colldb._colldb.crystal[name] is None:
                        line[name].material = SixTrack_to_xcoll[colldb.material[name]][0]
                    else:
                        line[name].material = SixTrack_to_xcoll[colldb.material[name]][1]
                if isinstance(line[name], EverestCrystal):
                    line[name].align_angle    = colldb._colldb.align_angle[name]
                    line[name].bending_radius = colldb._colldb.bending_radius[name]
                    line[name].xdim           = colldb._colldb.xdim[name]
                    line[name].ydim           = colldb._colldb.ydim[name]
                    line[name].thick          = colldb._colldb.thick[name]
                    line[name].miscut         = colldb._colldb.miscut[name]
                    line[name].lattice        = colldb._colldb.crystal[name]
            else:
                raise ValueError(f"Missing implementation for element type of collimator {name}!")
        colldb.gap = gaps_OLD

    @property
    def openings_set(self):
        # TODO: need to delete jaw positions if some parameters (that would influence it) are changed
        return not np.any(
                    [x is None for x in self.colldb._colldb.jaw_LU]
                    + [ x is None for x in self.colldb._colldb.jaw_RU]
                    + [ x is None for x in self.colldb._colldb.jaw_LD]
                    + [ x is None for x in self.colldb._colldb.jaw_RD]
                )


    def _generate_4D_pencil_one_jaw(self, num_particles, collimator, plane, side, impact_parameter,
                                    dr_sigmas, transverse_spread_sigma, match_at_s):

        line = self.line

        if plane == 'x':
            co_pencil     = line[collimator].ref_x
            co_transverse = line[collimator].ref_y
        else:
            co_pencil     = line[collimator].ref_y
            co_transverse = line[collimator].ref_x

        nemitt_x   = self.colldb.emittance[0]
        nemitt_y   = self.colldb.emittance[1]

        if side == '+':
            absolute_cut = line[collimator].jaw_LU + co_pencil + impact_parameter
        elif side == '-':
            absolute_cut = line[collimator].jaw_RU + co_pencil - impact_parameter

        # Collimator plane: generate pencil distribution
        pencil, p_pencil = xp.generate_2D_pencil_with_absolute_cut(num_particles,
                        plane=plane, absolute_cut=absolute_cut, dr_sigmas=dr_sigmas,
                        side=side, line=line, nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                        at_element=collimator, match_at_s=match_at_s
        )

        # Other plane: generate gaussian distribution in normalized coordinates
        transverse_norm   = np.random.normal(loc=co_transverse, scale=transverse_spread_sigma, size=num_particles)
        p_transverse_norm = np.random.normal(scale=transverse_spread_sigma, size=num_particles)

        return [pencil, p_pencil, transverse_norm, p_transverse_norm]


    def generate_pencil_on_collimator(self, collimator, num_particles, *, side='+-', impact_parameter=0, 
                                      pencil_spread=1e-6, transverse_impact_parameter=0.,
                                      transverse_spread_sigma=0.01, longitudinal=None,
                                      longitudinal_betatron_cut=None, sigma_z=7.61e-2):
        if not self.openings_set:
            raise ValueError("Need to set collimator openings before generating pencil distribution!")
        if not self.tracker_ready:
            raise Exception("Please build tracker before generating pencil distribution!")
        if transverse_impact_parameter != 0.:
            raise NotImplementedError

        # TODO: check collimator in colldb and installed!

        if self.colldb.side[collimator] == 'left':
            side = '+'
        if self.colldb.side[collimator] == 'right':
            side = '-'

        # Define the plane
        angle = self.colldb.angle[collimator]
        if abs(np.mod(angle-90,180)-90) < 1e-6:
            plane = 'x'
        elif abs(np.mod(angle,180)-90) < 1e-6:
            plane = 'y'
        else:
            raise NotImplementedError("Pencil beam on a skew collimator not yet supported!")

        nemitt_x   = self.colldb.emittance[0]
        nemitt_y   = self.colldb.emittance[1]

        # Is it converging or diverging?
        is_converging = self.colldb._optics.loc[self.s_active_front[collimator], 'alf' + plane ] > 0
        print(f"Collimator {collimator} is {'con' if is_converging else 'di'}verging.")
        if is_converging:
            # pencil at front of jaw
            match_at_s = self.s_active_front[collimator]
            sigma      = self.colldb._beam_size_front[collimator]
        else:
            # pencil at back of jaw
            match_at_s = self.s_active_back[collimator]
            sigma      = self.colldb._beam_size_back[collimator]

        dr_sigmas = pencil_spread/sigma

        # Generate 4D coordinates
        if side == '+-':
            num_plus = int(num_particles/2)
            num_min  = int(num_particles - num_plus)
            coords_plus = self._generate_4D_pencil_one_jaw(num_plus, collimator, plane, '+',
                                                           impact_parameter, dr_sigmas,
                                                           transverse_spread_sigma, match_at_s)
            coords_min  = self._generate_4D_pencil_one_jaw(num_min, collimator, plane, '-',
                                                           impact_parameter, dr_sigmas,
                                                           transverse_spread_sigma, match_at_s)
            coords      = [ [*c_plus, *c_min] for c_plus, c_min in zip(coords_plus, coords_min)]
        else:
            coords      = self._generate_4D_pencil_one_jaw(num_particles, collimator, plane, side,
                                                           impact_parameter, dr_sigmas,
                                                           transverse_spread_sigma, match_at_s)
        pencil            = coords[0]
        p_pencil          = coords[1]
        transverse_norm   = coords[2]
        p_transverse_norm = coords[3]

        # Longitudinal plane
        # TODO: make this more general, make this better
        if longitudinal is None:
            delta = 0
            zeta  = 0
        elif longitudinal == 'matched_dispersion':
            if longitudinal_betatron_cut is None:
                cut = 0
            else:
                cut = np.random.uniform(-longitudinal_betatron_cut, longitudinal_betatron_cut, num_particles)
            delta = self.generate_delta_from_dispersion(at_element=collimator, plane=plane, position_mm=pencil,
                                                        betatron_cut=cut)
            zeta  = 0
        elif longitudinal == 'bucket':
            zeta, delta = xp.generate_longitudinal_coordinates(
                    num_particles=num_particles, distribution='gaussian', sigma_z=sigma_z, line=self.line
            )
        elif not hasattr(longitudinal, '__iter__'):
            raise ValueError
        elif len(longitudinal) != 2:
            raise ValueError
        elif isinstance(longitudinal, str):
            raise ValueError
        elif isinstance(longitudinal, dict):
            zeta = longitudinal['zeta']
            delta = longitudinal['delta']
        else:
            zeta = longitudinal[0]
            delta = longitudinal[1]

        # Build the particles
        if plane == 'x':
            part = xp.build_particles(
                    x=pencil, px=p_pencil, y_norm=transverse_norm, py_norm=p_transverse_norm,
                    zeta=zeta, delta=delta, nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                    line=self.line, at_element=collimator, match_at_s=match_at_s,
                    _buffer=self._part_buffer
            )
        else:
            part = xp.build_particles(
                    x_norm=transverse_norm, px_norm=p_transverse_norm, y=pencil, py=p_pencil, 
                    zeta=zeta, delta=delta, nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                    line=self.line, at_element=collimator, match_at_s=match_at_s,
                    _buffer=self._part_buffer
            )

        part._init_random_number_generator()

        return part


    def generate_delta_from_dispersion(self, at_element, *, plane, position_mm, betatron_cut=0):
        line = self.line
        if line.tracker is None:
            raise ValueError("Need to build tracker first!")
        if not hasattr(betatron_cut, '__iter__'):
            if hasattr(position_mm, '__iter__'):
                betatron_cut = np.full_like(position_mm, betatron_cut)
        elif not hasattr(position_mm, '__iter__'):
            position_mm = np.full_like(betatron_cut, position_mm)
        elif len(position_mm) != len(betatron_cut):
            raise ValueError
        tw = line.twiss()
        if isinstance(at_element, str):
            idx = line.element_names.index(at_element)
        betagamma = line.particle_ref.beta0[0] * line.particle_ref.gamma0[0]
        if plane == 'x':
            sigma = np.sqrt(tw.betx[idx]*self.colldb.emittance[0]/betagamma)
            delta = (position_mm - betatron_cut*sigma - tw.x[idx]) / tw.dx[idx]
        elif plane == 'y':
            sigma = np.sqrt(tw.bety[idx]*self.colldb.emittance[1]/betagamma)
            delta = (position_mm - betatron_cut*sigma - tw.y[idx]) / tw.dy[idx]
        else:
            raise ValueError("The variable 'plane' needs to be either 'x' or 'y'!")
        return delta


    def enable_scattering(self):
        # Prepare collimators for tracking
        for el in self.line.get_elements_of_type(element_classes)[0]:
            el.enable_scattering()
        self._set_record_impacts()

    def disable_scattering(self):
        # Prepare collimators for tracking
        for el in self.line.get_elements_of_type(element_classes)[0]:
            el.disable_scattering()

    @property
    def scattering_enabled(self):
        all_enabled  = np.all([self.line[coll]._tracking if hasattr(self.line[coll], '_tracking')
                               else True for coll in self.collimator_names])
        some_enabled = np.any([self.line[coll]._tracking if hasattr(self.line[coll], '_tracking')
                               else True  for coll in self.collimator_names])
        if some_enabled and not all_enabled:
            raise ValueError("Some collimators are enabled for tracking but not all! "
                           + "This should not happen.")
        return all_enabled


    def summary(self, part, weights=None, show_zeros=False, file=None, recompute=False):

        # We cache the result
        if (self._summary is None or self._part is None
            or not xt.line._dicts_equal(part.to_dict(), self._part)
            or recompute
           ):
            if weights is None:
                weights = np.ones(len(part.x))
            else:
                part.sort(interleave_lost_particles=True)
            self._part   = part.to_dict()
            coll_mask    = (part.state<=-330) & (part.state>=-340)
            coll_losses  = np.array([self.line.element_names[i] for i in part.at_element[coll_mask]])
            coll_loss_single = np.unique(coll_losses)
            coll_lengths = [self.line[nn].active_length for nn in self.collimator_names] 
            coll_pos     = [self.colldb.s_center[nn]    for nn in self.collimator_names]
            if self._line_is_reversed:
                coll_pos = [self.machine_length - s for s in coll_pos ]
            coll_types   = [self.line[nn].__class__.__name__  for nn in self.collimator_names]
            coll_weights = weights[coll_mask]
            nabs         = [coll_weights[coll_losses==nn].sum() for nn in self.collimator_names]

            self._summary = pd.DataFrame({
                        "collname": self.collimator_names,
                        "nabs":     nabs,
                        "length":   coll_lengths,
                        "s":        coll_pos,
                        "type":     coll_types
                      })

        if file is not None:
            with open(Path(file), 'w') as fid:
                fid.write(self._summary.__repr__())

        if show_zeros:
            return self._summary
        else:
            return self._summary[self._summary.nabs > 0]


    def create_weights_from_initial_state(part, function):
        if len(function) == 4:
            return function[0](part.x)*function[1](part.px)*\
                   function[2](part.y)*function[3](part.py)
        elif len(function) == 6:
            return function[0](part.x)*function[1](part.px)*\
                   function[2](part.y)*function[3](part.py)*\
                   function[4](part.zeta)*function[5](part.delta)
        else:
            raise NotImplementedError


    def lossmap(self, part, interpolation=0.1, file=None, recompute=False, weights=None):

        # We cache the result
        if (self._lossmap is None or self._part is None
            or not xt.line._dicts_equal(part.to_dict(), self._part)
            or recompute
           ):

            self._part = part.to_dict()

            # Loss location refinement
            if interpolation is not None:
                # We need to correct for particles scattered out of the collimator beyond the aperture restriction:
                # TODO: this should be done in the scattering code!!!
                new_state = part.state
                new_elem  = part.at_element
                for idx, elem in enumerate(part.at_element):
                    if (part.state[idx]==0 and elem > 0 and
                        self.line.element_names[elem-1] in self.collimator_names
                       ):
                        print(f"Found at {self.line.element_names[elem]}, should be {self.line.element_names[elem-1]}")
                        new_state[idx] = -339 # lost on collimator, special state  ## TODO get the correct flag
                        new_elem[idx] = elem - 1
                part.state      = new_state
                part.at_element = new_elem
                # Do the interpolation
                aper_s = list(part.s[part.state==0])
                if len(aper_s) > 0:
                    print("Performing the aperture losses refinement.")
                    loss_loc_refinement = xt.LossLocationRefinement(self.line,
                            n_theta = 360, # Angular resolution in the polygonal approximation of the aperture
                            r_max = 0.5, # Maximum transverse aperture in m
                            dr = 50e-6, # Transverse loss refinement accuracy [m]
                            ds = interpolation, # Longitudinal loss refinement accuracy [m]
                            # save_refine_trackers=True # Diagnostics flag
                            )
                    loss_loc_refinement.refine_loss_location(part)

            aper_s, aper_names, aper_nabs = self._get_aperture_losses(part, weights)
            coll_summary = self.summary(part, weights, recompute=recompute, show_zeros=False).to_dict('list')

            self._lossmap = {
                'collimator': {
                    's':      coll_summary['s'],
                    'name':   coll_summary['collname'],
                    'length': coll_summary['length'],
                    'n':      coll_summary['nabs']
                }
                ,
                'aperture': {
                    's':    aper_s,
                    'name': aper_names,
                    'n':    aper_nabs
                }
                ,
                'machine_length': self.machine_length
                ,
                'interpolation': interpolation
                ,
                'reversed': self._line_is_reversed
            }

        if file is not None:
            with open(Path(file), 'w') as fid:
                json.dump(self._lossmap, fid, cls=xo.JEncoder, indent=True)
    
        return self._lossmap


    def _get_aperture_losses(self, part, weights=None):
        if weights is None:
            weights = np.ones(len(part.x))
        else:
            part.sort(interleave_lost_particles=True)

        # Get s position per particle (lost on aperture)
        aper_mask = part.state==0
        aper_s = list(part.s[aper_mask])
        if len(aper_s) == 0:
            return [], [], []
        if self._line_is_reversed:
            aper_s = [ self.machine_length - s for s in aper_s ]

        # Store names of aperture markers
        aper_names   = [self.line.element_names[i] for i in part.at_element[aper_mask]]
        name_dict    = dict(zip(aper_s, aper_names)) # TODO: not floating-point-safe and slow

        # Create output arrays
        aper_pos     = np.unique(aper_s)
        aper_weights = weights[aper_mask]
        aper_nabs    = [aper_weights[aper_s==ss].sum() for ss in aper_pos] # TODO: this might be slow
        aper_names   = [name_dict[ss] for ss in aper_pos]
        return aper_pos, aper_names, aper_nabs




# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

from .general import _pkg_root, __version__, citation

from .beam_elements import BlackAbsorber, EverestBlock, EverestCollimator, EverestCrystal, element_classes
from .scattering_routines.everest import materials, Material, CrystalMaterial
from .manager import CollimatorManager
from .colldb import CollimatorDatabase, load_SixTrack_colldb
from .rf_sweep import RFSweep

# print("If you use Xcoll in your simulations, please cite us :-)")
# print(citation)



import io
import json
import numpy as np
import pandas as pd


# in colldb:
#      (gap [sigma] + offset [m] + tilt [deg]) or opening [m]
#      angle
#      length (all three)
#      side
#      parking ??
#      material
#      stage
#      family
#      crystal

# in element:
#      jaw
#      ref
#      angle
#      length (all three)
#      side
#      material
#      is_active

# in neither (as they are sequence/session-dependent):
#      align_to
#      s_center
#      collimator_type

# if physical_opening is used, other variables like tilt, opening, offset, .. are ignored
_element_properties  = {
                        'active_length':    0,
                        'inactive_front':   0,
                        'inactive_back':    0,
                        'jaw':              None,
                        'reference_center': None,
                        'angle':            0,
                        'side':             'both',
                        'material':         None,
                        'active':           True,
                        'crystal':          False
                       }
_crystal_properties  = {
                        'bend':             None,
                        'xdim':             0,
                        'ydim':             0,
                        'miscut':           0,
                        'thick':            0,
#         'align_angle':    xo.Float64,  #  = - sqrt(eps/beta)*alpha*nsigma
#         'crytilt':        xo.Float64,
#         'orient':         xo.Float64,
                       }
_sequence_properties = {
                        's_center':         None,
                        'align_to':         None,
                        'collimator_type':  None,
                       }
_colldb_properties   = {
                        'gap':              None,  # [L, R]
                        'offset':           0,     # single value: shift of gap in mm
                        'extra_mm':         0,     # single value: widening of gap
                        'opening_mm':       None,  # [L, R]
                        'tilt':             0,
                        'parking':          1,
                        'stage':            None,
                        'family':           None,
                        'overwritten_keys': []
                       }
_optics_vals = ['x', 'px', 'y', 'py', 'betx', 'bety', 'alfx', 'alfy', 'dx', 'dy', 'sigma_x', 'sigma_y']


# This creates a view on the settings of one collimator, kept in sync with the main database
class CollimatorSettings:

    def __init__(self, name, colldb=None, optics=None, element=None):
        if colldb is None:
            colldb = pd.DataFrame({**_element_properties, **_crystal_properties}, index=[name])
        self._colldb = colldb
        if name not in colldb.index:
            raise ValueError(f"Collimator {name} not found in database!")
        self._optics  = optics
        self._element = element
        self._name = name
        self._jaws_manually_set = False

        # Automatically assign all properties from _collimator_properties
        # Now create @property's for each of them:
        for pp in set(_coll_properties.keys()).union(set(_crystal_properties.keys())):
            if not pp in self._colldb.columns:
                raise ValueError(f"Error in settings for {self.name}: "
                               + f"Could not find property {pp} in database column header!")
            if pp in _properties_no_setter:
                setattr(self.__class__, pp, property(_prop_fget(self, pp)))
            else:
                setattr(self.__class__, pp, property(_prop_fget(self, pp), _prop_fset(self, pp)))

    @property
    def name(self):
        return self._name

    def to_dict(self):
        props = list(_coll_properties.keys()) + list(_crystal_properties.keys()) + _add_to_dict
        all_properties = {pp: getattr(self,pp) for pp in props if pp not in _properties_no_setter}
        return {'name': self.name, **all_properties}

    # Some easy accessors to the LR / LRUD properties:
    # -----------------------------------------------

    @property
    def angle(self):
        return _get_LR(self, 'angle')

    @angle.setter
    def angle(self, val):
        _set_LR(self, 'angle', val)
        self._compute_jaws()

    @property
    def tilt(self):
        return _get_LR(self, 'tilt')

    @tilt.setter
    def tilt(self, val):
        _set_LR(self, 'tilt', val)
        self._compute_jaws()

    @property
    def opening(self):
        if self._jaws_manually_set:
            # TODO: update gap using beam size as from jaws
            pass
        else:
            return _get_LR(self, 'gap')

    @opening.setter
    def opening(self, val):
        _set_LR(self, 'gap', val, neg=True)
        self._jaws_manually_set = False
        self._compute_jaws()

    @property
    def physical_opening(self):
        return _get_LRUD(self, 'jaw', neg=True)

    @physical_opening.setter
    def physical_opening(self, val):
        _set_LRUD(self, 'jaw', val, neg=True)
        self._jaws_manually_set = True
        self._compute_jaws()

    @property
    def reference_center(self):
        return _get_LRUD(self, 'ref', name_LU='_xU', name_RU='_yU', name_LD='_xD', name_RD='_yD')

    @property
    def beam_size(self):
        if not self._optics_is_ready:
            return None
        else:
            # TODO: calculate beam size from optic
            # [[LU,RU], [LC, RC], [LD,RD]]
            pass


    @property
    def _optics_is_ready(self):
        if self.align_to is None or self._optics is None:
            return False
        # TODO: when True

    def _compute_jaws(self):
        if self._optics_is_ready:
            if self._jaws_manually_set:
                jaw_LU = self.jaw_LU
                jaw_RU = self.jaw_RU
                jaw_LD = self.jaw_LD
                jaw_RD = self.jaw_RD
                # TODO: upate gap
                pass
            else:
                # Default to parking
                # TODO: parking is wrt CO, need to correct
                jaw_LU = self.parking
                jaw_RU = -self.parking
                jaw_LD = self.parking
                jaw_RD = -self.parking

                if self.gap_L is not None or self.gap_R is not None:
                    # Get the beam size to be used, depending on align_to
                    if self.align_to == 'maximum':
                        # Reset align_to to the location of the maximum
                        self.align_to = ['front', 'center', 'back'][np.argmax(self.beam_size)]
                    if self.align_to == 'angular':
                        sigma_U = self.beam_size[0]
                        sigma_D = self.beam_size[2]
                    elif self.align_to == 'front':
                        sigma_U = self.beam_size[0]
                        sigma_D = self.beam_size[0]
                    elif self.align_to == 'center':
                        sigma_U = self.beam_size[1]
                        sigma_D = self.beam_size[1]
                    elif self.align_to == 'back':
                        sigma_U = self.beam_size[2]
                        sigma_D = self.beam_size[2]

                    # Get the shift due to the tilt to be used, depending on align_to
                    # TODO: is this correct?
                    # TODO: better name for align_to?
                    if align_to == front:
                        scale = 0
                    elif align_to == back:
                        scale = 1
                    else:
                        scale = 0.5
                    ts_LU = -np.tan(self.tilt_L)*self.active_length*scale
                    ts_RU = -np.tan(self.tilt_R)*self.active_length*scale
                    ts_LD = np.tan(self.tilt_L)*self.active_length*(1-scale)
                    ts_RD = np.tan(self.tilt_R)*self.active_length*(1-scale)

                    if self.side in ['left', 'both']:
                        jaw_LU = self.gap_L*sigma_U[0]  + self.offset + ts_LU
                        jaw_LD = self.gap_L*sigma_D[0]  + self.offset + ts_LD
                    if self.side in ['right', 'both']:
                        jaw_RU = -self.gap_R*sigma_U[1] + self.offset + ts_RU
                        jaw_RD = -self.gap_R*sigma_D[1] + self.offset + ts_RD

            if self.side == 'left':
                self.jaw_LU = min(jaw_LU, self.parking)
                self.jaw_LD = min(jaw_LD, self.parking)
                self.jaw_RU = None
                self.jaw_RD = None
            elif self.side == 'right':
                self.jaw_LU = None
                self.jaw_LD = None
                self.jaw_RU = max(jaw_RU, -self.parking)
                self.jaw_RD = max(jaw_RD, -self.parking)
            else:
                self.jaw_LU = min(jaw_LU, self.parking)
                self.jaw_LD = min(jaw_LD, self.parking)
                self.jaw_RU = max(jaw_RU, -self.parking)
                self.jaw_RD = max(jaw_RD, -self.parking)


# Helper functions to set/get properties
# --------------------------------------

def _get_LR(obj, prop, neg=False, name_L='_L', name_R='_R'):
    # Is the property reflected along left/right?
    sign = -1 if neg else 1
    # Is it a property or a dict key?
    if isinstance(obj, dict):
        L = obj[prop + name_L]
        R = obj[prop + name_R]
    else:
        L = getattr(obj, prop + name_L)
        R = getattr(obj, prop + name_R)
    # Find out how many values to return
    if L is None and R is None:
        return None
    elif L is None:
        return R
    elif R is None:
        return L
    else:
        return L if L==sign*R else [L,R]


def _set_LR(obj, prop, val, neg=False, name=None, name_L='_L', name_R='_R'):
    # 'name' is only used for error reporting
    if name is None:
        if isinstance(obj, dict):
            name = 'dict_property'
        else:
            name = obj.name if hasattr(obj, 'name') else obj.__class__.__name__
    # Is the property reflected along left/right?
    sign = -1 if neg else 1
    # Find out how to set values
    if not hasattr(val, '__iter__'):
        val = [val]
    error_dimension = False
    if isinstance(val, str):
        raise ValueError(f"Error in settings for {name}: "
                       + f"The setting `{prop}` has to be a number!")
    elif len(val) == 2:
    # The value is of the form [val_L,val_R]
        if hasattr(val[0], '__iter__') or hasattr(val[1], '__iter__'):
            error_dimension = True
        val_L = val[0]
        val_R = val[1]
    elif len(val) == 1:
    # The value is of the form [val]
        if hasattr(val[0], '__iter__'):
            error_dimension = True
        val_L = val[0]
        val_R = sign*val[0]
    else:
        error_dimension = True
    if error_dimension:
        raise ValueError(f"Error in settings for {name}: "
                       + f"The setting `{prop}` must have one or two (L, R) values!")
    # Is it a property or a dict key?
    if isinstance(obj, dict):
        obj[prop + name_L] = val_L
        obj[prop + name_R] = val_R
    else:
        _prop_fset(obj, prop + name_L)(obj, val_L)
        _prop_fset(obj, prop + name_R)(obj, val_R)


def _get_LRUD(obj, prop, neg=False, name=None,
              name_LU='_LU', name_RU='_RU', name_LD='_LD', name_RD='_RD'):
    # 'name' is only used for error reporting
    if name is None:
        if isinstance(obj, dict):
            name = 'dict_property'
        else:
            name = obj.name if hasattr(obj, 'name') else obj.__class__.__name__
    # Is the property reflected along left/right?
    sign = -1 if neg else 1
    # Is it a property or a dict key?
    if isinstance(obj, dict):
        LU = obj[prop + name_LU]
        RU = obj[prop + name_RU]
        LD = obj[prop + name_LD]
        RD = obj[prop + name_RD]
    else:
        LU = getattr(obj, prop + name_LU)
        RU = getattr(obj, prop + name_RU)
        LD = getattr(obj, prop + name_LD)
        RD = getattr(obj, prop + name_RD)
    # Find out how many values to return
    if LU is None and RU is None \
    and LD is None and RD is None:
        return None
    elif LU is None and RU is None:
        return LD if LD==sign*RD else [LD,RD]
    elif LD is None and RD is None:
        return LU if LU==sign*RU else [LU,RU]
    elif LU is None and RD is None:
        raise ValueError(f"Error in settings for {name}: "
                       + f"The setting `{prop}` has values for LD and RU but not for LU and RD."
                       + f"Cannot mix jaws L/R with U/D! Either set all four, or only L or only R.")
    elif LD is None and RU is None:
        raise ValueError(f"Error in settings for {name}: "
                       + f"The setting `{prop}` has values for LU and RD but not for LD and RU."
                       + f"Cannot mix jaws L/R with U/D! Either set all four, or only L or only R.")
    else:
        if LU == sign*RU == LD == sign*RD:
            return LU
        elif LU == LD and RU == RD:
            return [LU, RU]
        else:
            return [[LU, RU], [LD, RD]]


def _set_LRUD(obj, prop, val, neg=False, name=None,
              name_LU='_LU', name_RU='_RU', name_LD='_LD', name_RD='_RD'):
    # 'name' is only used for error reporting
    if name is None:
        if isinstance(obj, dict):
            name = 'dict_property'
        else:
            name = obj.name if hasattr(obj, 'name') else obj.__class__.__name__
    # Is the property reflected along left/right?
    sign = -1 if neg else 1

    # Find out how to set values
    if not hasattr(val, '__iter__'):
        val = [val]
    error_string    = False
    error_dimension = False
    if isinstance(val, str):
        error_string = True
    elif len(val) == 4:
    # The value is of the form [val_LU,val_RU,val_LD,val_RD]
        if hasattr(val[0], '__iter__') or hasattr(val[1], '__iter__') \
        or hasattr(val[2], '__iter__') or hasattr(val[3], '__iter__'):
            error_dimension = True
        val_LU = val[0]
        val_RU = val[1]
        val_LD = val[2]
        val_RD = val[3]
    elif len(val) == 2:
        if not hasattr(val[0], '__iter__') and not hasattr(val[1], '__iter__'):
        # The value is of the form [val_L,val_R]
            val_LU = val[0]
            val_RU = val[1]
            val_LD = val[0]
            val_RD = val[1]
        elif not hasattr(val[0], '__iter__') or not hasattr(val[1], '__iter__'):
            error_dimension = True
        else:
        # The value is of the form [[val_LU, val_RU], [val_LD, val_RD]]
            if isinstance(val[0], str) or isinstance(val[1], str):
                error_string = True
            if hasattr(val[0][0], '__iter__') or hasattr(val[0][1], '__iter__') \
            or hasattr(val[1][0], '__iter__') or hasattr(val[1][1], '__iter__'):
                error_dimension = True
            val_LU = val[0][0]
            val_RU = val[0][1]
            val_LD = val[1][0]
            val_RD = val[1][1]
    elif len(val) == 1:
    # The value is of the form [val]
        if hasattr(val[0], '__iter__'):
            error_dimension = True
        val_LU = val[0]
        val_RU = sign*val[0]
        val_LD = val[0]
        val_RD = sign*val[0]
    else:
        error_dimension = True
    if error_dimension:
        raise ValueError(f"Error in settings for {name}: "
                       + f"The setting `{prop}` has to be given as val, [val], "
                       + f"[val_L, val_R], [[val_LU, val_RU], [val_LD, val_RD]], "
                       + f"or [val_LU, val_RU, val_LD, val_RD]!")
    if error_string:
        raise ValueError(f"Error in settings for {name}: "
                               + f"The setting `{prop}` has to be a number or a "
                               + f"list of numbers!")

    # Is it a property or a dict key?
    if isinstance(obj, dict):
        obj[prop + name_LU] = val_LU
        obj[prop + name_RU] = val_RU
        obj[prop + name_LD] = val_LD
        obj[prop + name_RD] = val_RD
    else:
        _prop_fset(obj, prop + name_LU)(obj, val_LU)
        _prop_fset(obj, prop + name_RU)(obj, val_RU)
        _prop_fset(obj, prop + name_LD)(obj, val_LD)
        _prop_fset(obj, prop + name_RD)(obj, val_RD)


# These getter and setter functions link each property to the corresponding
# entry in the DataFrame
def _prop_fget(obj, attr_name):
    def fget(obj):
        return obj._colldb.loc[obj.name, attr_name]
    return fget

def _prop_fset(obj, attr_name):
    def fset(obj, value):
        if hasattr(obj, '_colldb'):
            obj._colldb.loc[obj.name, attr_name] = value
        if hasattr(obj, attr_name):
            setattr(obj, attr_name, value)
        # If we want additional effects on the setter funcions, this
        # can be achieved by defining an fset_prop method
        if hasattr(obj.__class__, 'fset_' + attr_name):
            getattr(obj, 'fset_' + attr_name)(value)
        # TODO: self._compute_jaws()
    return fset




# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import io
import json
import numpy as np
import pandas as pd


#TODO: niet-active collimators op non-active etc (ook crystals)

def load_SixTrack_colldb(filename, *, emit):
    print("Warning: Using 'xcoll.load_SixTrack_colldb()' is deprecated! "
        + "Use 'xcoll.CollimatorDatabase.from_SixTrack()' instead.")
    return CollimatorDatabase.from_SixTrack(file=filename, nemitt_x=emit, nemitt_y=emit)


def _initialise_None(collimator):
    fields = {'s_center':None, 'align_to': None, 's_align_front': None, 's_align_back': None }
    fields.update({'gap_L': None, 'gap_R': None, 'angle_L': 0, 'angle_R': 0, 'offset': 0, 'parking': 1})
    fields.update({'jaw_LU': None, 'jaw_RU': None, 'jaw_LD': None, 'jaw_RD': None, 'family': None, 'overwritten_keys': []})
    fields.update({'side': 'both', 'material': None, 'stage': None, 'collimator_type': None, 'active': True})
    fields.update({'active_length': 0, 'inactive_front': 0, 'inactive_back': 0, 'sigmax': None, 'sigmay': None})
    fields.update({'crystal': None, 'bending_radius': None, 'xdim': 0, 'ydim': 0, 'miscut': 0, 'thick': 0})
    for f, val in fields.items():
        if f not in collimator.keys():
            collimator[f] = val
    for key in collimator.keys():
        if key not in fields.keys():
            raise ValueError(f"Illegal setting {key} in collimator!")


def _dict_keys_to_lower(dct):
    if isinstance(dct, dict):
        return {k.lower(): _dict_keys_to_lower(v) for k,v in dct.items()}
    else:
        return dct


def _get_coll_dct_by_beam(coll, beam):
    # The dictionary can be a CollimatorDatabase for a single beam (beam=None)
    # or for both beams (beam='b1' or beam='b2)
    if beam is not None:
        if isinstance(beam, int) or len(beam) == 1:
            beam = f'b{beam}'
        beam = beam.lower()
    beam_in_db = list(coll.keys())
    
    if beam_in_db == ['b1','b2']:
        if beam is None:
            raise ValueError("Need to specify a beam, because the given dict is for both beams!")
        return coll[beam]

    elif len(beam_in_db) == 1:
        if beam is None:
            beam = beam_in_db[0].lower()
        elif beam != beam_in_db[0].lower():
            raise ValueError("Variable 'beam' does not match beam specified in CollimatorDatabase!")
        return coll[beam]

    elif beam is not None:
        print("Warning: Specified a beam, but the CollimatorDatabase is for a single beam only!")
    return coll


class CollimatorDatabase:

    _init_vars = ['collimator_dict', 'family_dict', 'beam', 'nemitt_x', 'nemitt_y', '_yaml_merged', 'ignore_crystals']
    _init_var_defaults = {'family_dict': {}, 'beam': None, '_yaml_merged': False, 'ignore_crystals': True}

    # -------------------------------
    # ------ Loading functions ------
    # -------------------------------
    
    @classmethod
    def from_yaml(cls, file, **kwargs):

        # Only do the import here, as to not force people to install
        # ruamel if they don't load CollimatorDatabase yaml's
        from ruamel.yaml import YAML
        yaml = YAML(typ='safe')
        if isinstance(file, io.IOBase):
            dct = yaml.load(file)
        else:
            with open(file, 'r') as fid:
                dct = yaml.load(fid)
        dct = _dict_keys_to_lower(dct)

        # If the CollimatorDatabase uses YAML merging, we need a bit of hackery to get the
        # family names from the tags (anchors/aliases)
        _yaml_merged = False
        if 'families' in dct.keys() and not isinstance(dct['families'], dict):
            _yaml_merged = True

            # First we load a round-trip yaml
            yaml = YAML()
            if isinstance(file, io.IOBase):
                full_dct = yaml.load(file)
            else:
                with open(file, 'r') as fid:
                    full_dct = yaml.load(fid)
            famkey = [key for key in full_dct.keys() if key.lower() == 'families'][0]
            collkey = [key for key in full_dct.keys() if key.lower() == 'collimators'][0]
            families = {}

            # We loop over family names to get the name tag ('anchor') of each family
            for fam, full_fam in zip(dct['families'], full_dct[famkey]):
                if full_fam.anchor.value is None:
                    raise ValueError("Missing name tag / anchor in "
                                   + "CollimatorDatabase['families']!")
                # We get the anchor from the rt yaml, and use it as key in the families dict
                families[full_fam.anchor.value.lower()] = {f.lower(): v for f, v in fam.items()}
            dct['families'] = families

            # Now we need to loop over each collimator, and verify which family was used
            beam = kwargs.get('beam', None)
            coll_dct = _get_coll_dct_by_beam(dct['collimators'], beam)
            full_coll_dct = _get_coll_dct_by_beam(full_dct[collkey], beam)
            for coll, full_coll in zip(coll_dct.values(), full_coll_dct.values()):
                if 'family' in coll.keys():
                    raise ValueError(f"Error in {coll}: Cannot use merging for families "
                                    + "and manually specify family as well!")
                elif len(full_coll.merge) > 0:
                    coll['family'] = full_coll.merge[0][1].anchor.value.lower()
                    # Check if some family settings are overwritten for this collimator
                    overwritten_keys = [key.lower() for key in full_coll.keys()
                                        if full_coll._unmerged_contains(key)
                                        and key.lower() in families[coll['family']].keys()]
                    if len(overwritten_keys) > 0:
                        coll['overwritten_keys'] = overwritten_keys
                else:
                    coll['family'] = None

        return cls.from_dict(dct, _yaml_merged=_yaml_merged, **kwargs)


    @classmethod
    def from_json(cls, file, **kwargs):
        if isinstance(file, io.IOBase):
            dct = json.load(file)
        else:
            with open(file, 'r') as fid:
                dct = json.load(fid)
        dct = _dict_keys_to_lower(dct)
        return cls.from_dict(dct, **kwargs)


    @classmethod
    def from_dict(cls, dct, beam=None, _yaml_merged=False, nemitt_x=None, nemitt_y=None, ignore_crystals=True):
        # We make all keys case-insensitive to avoid confusion between different conventions
        # The families are optional
        fam = {}
        dct = _dict_keys_to_lower(dct)

        # Get the emittance
        if nemitt_x is None and nemitt_y is None:
            if 'emittance' not in dct.keys():
                raise ValueError("Missing emittance info! Add 'emittance' as a key to "
                               + "the CollimatorDatabase file, or specify it as 'nemitt_x' "
                               + "and 'nemitt_y' to the loader!")
            nemitt_x = dct['emittance']['x']
            nemitt_y = dct['emittance']['y']
        elif nemitt_x is None or nemitt_y is None:
            raise ValueError("Need to provide both 'nemitt_x' and 'nemitt_y'!")
        elif 'emittance' in dct.keys():
            if dct['emittance']['x'] != nemitt_x or dct['emittance']['y'] != nemitt_y:
                raise ValueError("Emittance in CollimatorDatabase file different from "
                               + "'nemitt_x' and 'nemitt_y'!")

        # Get family and collimator dicts
        if 'families' in dct.keys():
            if not 'collimators' in dct.keys():
                raise ValueError("Could not find 'collimators' dict in CollimatorDatabase!")
            fam  = dct['families']
            coll = dct['collimators']
        elif 'collimators' in dct.keys():
            coll = dct['collimators']
        else:
            coll = dct

        return cls(collimator_dict=coll, family_dict=fam, nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                   beam=beam, _yaml_merged=_yaml_merged, ignore_crystals=ignore_crystals)


    @classmethod
    def from_SixTrack(cls, file, ignore_crystals=True, **kwargs):
        # only import regex here
        import re
        with open(file, 'r') as infile:
            coll_data_string = ''
            family_settings = {}
            family_types = {}
            side = {}
            cry_fields = ['bending_radius', 'xdim', 'ydim', 'thick', 'miscut', 'crystal']
            cry = {}

            for l_no, line in enumerate(infile):
                if line.startswith('#'):
                    continue # Comment
                sline = line.split()
                if len(sline) > 0:
                    if sline[0].lower() == 'nsig_fam':
                        family_settings[sline[1]] = float(sline[2])
                        family_types[sline[1]] = sline[3]
                    elif sline[0].lower() == 'onesided':
                        side[sline[1]] = int(sline[2])
                    elif sline[0].lower() == "crystal":
                        cry[sline[1]] = {}
                        for i, key in enumerate(cry_fields):
                            idx = i+2 if i < 4 else i+3  # we skip "tilt"
                            if i < 5:
                                cry[sline[1]][key] = float(sline[idx])
                            else:
                                cry[sline[1]][key] = int(sline[idx])
                    elif sline[0].lower() == 'settings':
                        pass # Acknowledge and ignore this line
                    elif len(sline) == 6:
                        # Standard collimator definition
                        coll_data_string += line
                    else:
                        print(f"Unknown setting {line}")

        defaults = {}
        _initialise_None(defaults)

        famdct = {key: {'gap': family_settings[key], 'stage': family_types[key]} for key in family_settings}
        names = ['name', 'gap', 'material', 'active_length', 'angle', 'offset']

        df = pd.read_csv(io.StringIO(coll_data_string), sep=r'\s+', index_col=False, names=names)
        df['family'] = df['gap'].copy()
        df['family'] = df['family'].apply(lambda s: None if re.match(r'^-?\d+(\.\d+)?$', str(s)) else s)
        df.insert(5,'stage', df['gap'].apply(lambda s: family_types.get(s, 'UNKNOWN')))

        df['gap'] = df['gap'].apply(lambda s: float(family_settings.get(s, s)))
        # TODO this breaks code if a key has upper case, e.g. gap_L
        df['name'] = df['name'].str.lower() # Make the names lowercase for easy processing
        df['parking'] = 0.025
        if ignore_crystals:
            df = df[~df.name.isin(list(cry.keys()))]
        else:
            for key in cry_fields:
                df[key] = [cry[name][key] if name in cry else defaults[key]
                           for name in df['name']]
            df['crystal'] = ['strip' if s==1 else s for s in df['crystal']]
            df['crystal'] = ['quasi-mosaic' if s==2 else s for s in df['crystal']]
        df['side'] = [side[name] if name in side else defaults['side']
                      for name in df['name']]
        df['side'] = ['both'  if s==0 else s for s in df['side']]
        df['side'] = ['left'  if s==1 else s for s in df['side']]
        df['side'] = ['right' if s==2 else s for s in df['side']]
        df = df.set_index('name')

        return cls.from_dict({'collimators': df.transpose().to_dict(), 'families': famdct}, \
                             ignore_crystals=ignore_crystals, **kwargs)


    def write_to_yaml(self, out, lhc_style=True):
        """
        Writes a colldb in memory to disk in the yaml format.

        > colldb_object.write_to_yaml(<path+name>, lhc_style=Bool)

        if lhc_style == True, it will add comments assuming that the collimators are named
        as in the lhc.

        The function can dump b1, b2 and a general bx, however multi-beam functionality is not yet
        added to the collmanager. TODO

        If any of the dumped keys contains capital letters (e.g. gap_L), it will not be possible
        to load it back into xcoll, since all keys are set to lowercase when importing TODO
        """
        # Dumps collimator database to a YAML file with optional LHC style formatting
        import re

        # Local helper functions
        def _format_dict_entry(key, value, spacing='', mapping=False, key_width=15):
            # Formats a dictionary entry into a string for YAML output
            formatted_values = ',    '.join(f"{k}: {v}" for k, v in value.items())
            formatted_values = re.sub(r'none', 'null', formatted_values, flags=re.IGNORECASE)
            # Ensure key has a fixed width for alignment
            if mapping:
                formatted_key = f'{key}'.ljust(key_width)
            else:
                formatted_key = f'{key}:'.ljust(key_width)
            #formatted_values = formatted_values.ljust(key_width)
            return f"{spacing}{formatted_key} {{ {formatted_values} }}\n"
        
        def _print_values(keys, dct, file, spacing='', mapping=False):
            # Writes formatted dictionary entries to a file
            for key in keys:
                file.write(_format_dict_entry(key, dct[key], spacing=spacing, mapping=mapping))
        
        def _print_colls(colls, dcts, beam, file):
            # Filters and formats collimator data, then writes to a file
            coll_items_to_print = ['<<','gap','angle','material','active','length','side']
            file.write(f'  {beam}:\n')
            for coll in colls:
                coll_dict = dcts._colldb.transpose().to_dict()[coll]
                fam = coll_dict['family']
                fam_keys = []
                if fam is not None:
                    fam_keys = dcts._family_dict[fam].keys()
                    coll_dict = {**{'<<': '*'+fam}, **coll_dict}
                temp_items_to_print = []
                if coll_dict['crystal'] and str(coll_dict['crystal'])!='nan':
                    temp_items_to_print = ['bending_radius','xdim','ydim','miscut','crystal', 'thick']
                if coll_dict['angle_L'] == coll_dict['angle_R']:
                    coll_dict.update({'angle': coll_dict['angle_L']})
                else:
                    temp_items_to_print = temp_items_to_print + ['angle_L','angle_R']
                if coll_dict['gap_L'] == coll_dict['gap_R']:
                    coll_dict.update({'gap': coll_dict['gap_L']})
                elif coll_dict['gap_L'] is None and coll_dict['gap_R'] is not None:
                    coll_dict.update({'gap': coll_dict['gap_R']})
                elif coll_dict['gap_L'] is not None and coll_dict['gap_R'] is None:
                    coll_dict.update({'gap': coll_dict['gap_L']})
                else:
                    temp_items_to_print = temp_items_to_print + ['gap_L','gap_R']
                value = {}
                overwritten_keys = coll_dict['overwritten_keys']
                for key, val in coll_dict.items():
                    if key == 'active_length':
                        key = 'length' 
                    if (key in coll_items_to_print+temp_items_to_print) and (key not in (set(fam_keys)-set(overwritten_keys))) and (val != 'both'):
                        value.update({key: val})
                file.write(_format_dict_entry(coll, value, spacing='    '))
            file.write('\n')   
                                

        LHC_families = ['tcp3', 'tcsg3', 'tcsm3', 'tcla3', 'tcp7', 'tcsg7', 'tcsm7', 'tcla7', 'tcli', 'tdi', 'tcdq', 'tcstcdq', 'tcth1', 'tcth2', 'tcth5', 'tcth8', 'tctv1', 'tctv2', 'tctv5', 'tctv8', 'tclp', 'tcxrp', 'tcryo', 'tcl4', 'tcl5', 'tcl6', 'tct15', 'tct2', 'tct8', 'tcsp', 'tcld']
        with open(f'{out}.yaml', 'w') as file:
            if '_family_dict' in self.__dict__.keys():
                file.write('families:\n')
                if lhc_style:
                    printed_families = []
                    fams_in_dict = self._family_dict.keys()

                    # Momentum cleaning
                    file.write('  # Momentum cleaning\n')
                    sel_fam = [fam for fam in LHC_families if re.match('.*3', fam) and (fam in fams_in_dict)]
                    printed_families += sel_fam
                    _print_values(sel_fam, self._family_dict, file, spacing='  - &', mapping=True)

                    # Betatron cleaning
                    file.write('  # Betatron cleaning\n')
                    sel_fam = [fam for fam in LHC_families if re.match('.*7', fam) and (fam in fams_in_dict)]
                    printed_families += sel_fam
                    _print_values(sel_fam, self._family_dict, file, spacing='  - &', mapping=True)

                    # Injection protection
                    file.write('  # Injection protection\n')
                    sel_fam = [fam for fam in LHC_families if (fam in ['tcli', 'tdi']) and (fam in fams_in_dict)]
                    printed_families += sel_fam
                    _print_values(sel_fam, self._family_dict, file, spacing='  - &', mapping=True)

                    # Dump protection
                    file.write('  # Dump protection\n')
                    sel_fam = [fam for fam in LHC_families if (fam in ['tcdq', 'tcsp', 'tcstcdq']) and (fam in fams_in_dict)]
                    printed_families += sel_fam
                    _print_values(sel_fam, self._family_dict, file, spacing='  - &', mapping=True)

                    # Physics background / debris
                    file.write('  # Physics background / debris\n')
                    sel_fam = [fam for fam in LHC_families if ((re.match('tc[lt][0-9dp].*', fam)) or (fam in ['tcryo', 'tcxrp'])) and (fam in fams_in_dict)]
                    printed_families += sel_fam
                    _print_values(sel_fam, self._family_dict, file, spacing='  - &', mapping=True)

                    # Other families
                    if set(printed_families) != set(fams_in_dict):
                        file.write('  # Other families\n')
                        _print_values(set(fams_in_dict) - set(printed_families), self._family_dict, file, spacing='  - &', mapping=True)
                else:
                    file.write('  # Families\n')
                    _print_values(self._family_dict.keys(), self._family_dict, file, spacing='  - &', mapping=True)

            # Emittance section
            ex = self.emittance[0]
            ey = self.emittance[1]
            file.write(f'\nemittance:\n  x: {ex}\n  y: {ey}\n')

            # Collimators section
            file.write('\ncollimators:\n')
            b1_colls, b2_colls, bx_colls = [], [], []
            for coll in self._colldb.index:
                if coll == 'tclia.4r2' or coll == 'tclia.4l8':
                    b1_colls.append(coll)
                    b2_colls.append(coll)
                elif coll[-2:] == 'b1':
                    b1_colls.append(coll)
                elif coll[-2:] == 'b2':
                    b2_colls.append(coll)
                else:
                    bx_colls.append(coll)
            
            # Handle special cases for collimators
            if (('tclia.4r2' in b1_colls) or ('tclia.4l8' in b1_colls)) and (len(b1_colls) <= 2):
                b1_colls = []
            if (('tclia.4r2' in b2_colls) or ('tclia.4l8' in b2_colls)) and (len(b2_colls) <= 2):
                b2_colls = []

            # Print collimators for each beam
            if len(b1_colls) > 0:
                _print_colls(b1_colls, self, 'b1', file)
            if len(b2_colls) > 0:
                _print_colls(b2_colls, self, 'b2', file)
            if len(bx_colls) > 0:
                _print_colls(bx_colls, self, 'bx', file)
                print('WARNING -- some collimators could not be assigned to b1 or b2. Tracking might not work with those collimators. Please manually change the output file if necessary.')


    def __init__(self, **kwargs):
        # Get all arguments
        for var in self._init_vars:
            if var in self._init_var_defaults:
                kwargs.setdefault(var, self._init_var_defaults[var])
            elif var not in kwargs.keys():
                raise ValueError(f"CollimatorDatabase is missing required argument '{var}'!")

        self._optics = pd.DataFrame(columns=['x', 'px', 'y', 'py', 'betx', 'bety', 'alfx', 'alfy', 'dx', 'dy'])
        self._parse_dict(kwargs['collimator_dict'], kwargs['family_dict'],
                         kwargs['beam'], kwargs['_yaml_merged'], kwargs.get('ignore_crystals', True))
        self.emittance = [kwargs['nemitt_x'], kwargs['nemitt_y']]
        self._beta_gamma_rel = None


    def __getitem__(self, name):
        return CollimatorSettings(name, self._colldb)

    def to_pandas(self):
        return pd.DataFrame({
                's_center':        self.s_center,
                'gap':             self.gap,
                'jaw':             self.jaw,
                'beam_size':       self.beam_size,
                'aligned_to':      self.align_to,
                'angle':           self.angle,
                'material':        self.material,
#                 'offset':          self.offset,
#                 'tilt':            self.tilt,
                'stage':           self.stage,
                'active_length':   self.active_length,
                'collimator_type': self.collimator_type,
            }, index=self.name)


    def _parse_dict(self, coll, fam, beam=None, _yaml_merged=False, ignore_crystals=True):

        # We make all keys case-insensitive to avoid confusion between different conventions
        coll = _dict_keys_to_lower(coll)
        fam  = _dict_keys_to_lower(fam)

        # The dictionary can be a CollimatorDatabase for a single beam (beam=None)
        # or for both beams (beam='b1' or beam='b2)
        coll = _get_coll_dct_by_beam(coll, beam)

        # Apply family settings
        crystals = []
        for thiscoll, settings in coll.items():
            settings = {k.lower(): v for k,v in settings.items()}
            if 'family' in settings.keys() and settings['family'] is not None:
                settings['family'] = settings['family'].lower()
                thisfam = settings['family']
                if thisfam not in fam.keys():
                    raise ValueError(f"Collimator {thiscoll} depends on family {thisfam}, "
                                   + f"but the latter is not defined!")

                # Check if some family settings are overwritten for this collimator
                # Only do this check if we didn't do a YAML merge earlier (because then it
                # is already taken care of)
                if not _yaml_merged:
                    overwritten_keys = [key.lower() for key in settings.keys() if key in fam[thisfam]]
                    if len(overwritten_keys) > 0:
                        settings['overwritten_keys'] = overwritten_keys

                # Load family settings, potentially overwriting settings for this collimator
                settings = {**fam[thisfam], **settings}

            else:
                settings['family'] = None
            coll[thiscoll] = settings

            # Save list of crystals
            if 'crystal' in settings:
                if settings['crystal'] != 0.0:
                    crystals += [thiscoll]
                else: 
                    settings['crystal'] = None

        # Remove crystals from colldb
        if ignore_crystals:
            for thiscoll in crystals:
                del coll[thiscoll]

        # Check that all collimators have gap settings
        if not np.all(['gap' in val.keys() or 'opening' in val.keys() for val in coll.values()]):
            raise ValueError("Ill-defined CollimatorDatabase: Not all collimators have a gap or "
                           + "opening setting, (or the keys / structure of the dictionary is wrong)!")

        # Update collimators with default values for missing keys
        for collimator in coll.values():
            # Change all values to lower case
            for key, val in collimator.items():
                collimator[key] = val.lower() if isinstance(val, str) else val
            if 'length' in collimator.keys():
                collimator['active_length'] = collimator.pop('length')
            if 'gap' in collimator.keys():
                if collimator['gap'] is not None and collimator['gap'] > 900:
                    collimator['gap'] = None
                if 'side' in collimator.keys() and collimator['side'] == 'left':
                    collimator['gap_L'] = collimator.pop('gap')
                    collimator['gap_R'] = None
                elif 'side' in collimator.keys() and collimator['side'] == 'right':
                    collimator['gap_L'] = None
                    collimator['gap_R'] = collimator.pop('gap')
                else:
                    collimator['gap_L'] = collimator['gap']
                    collimator['gap_R'] = collimator.pop('gap')
            if 'angle' in collimator.keys():
                collimator['angle_L'] = collimator['angle']
                collimator['angle_R'] = collimator.pop('angle')
            _initialise_None(collimator)

        self._collimator_dict = coll
        self._family_dict = fam
        self._colldb = pd.DataFrame(coll).transpose()


    @property
    def name(self):
        return self._colldb.index.values


    # TODO: - VALIDATION OF TYPES (e.g. material, stage, align, ..)
    #       - IMPLEMENTATION OF TILT
    #       - CRYSTAL PROPERTIES: only valid if crystal == True (add mask to _set_property?)
    #         make second dataframe for crystals
    #       - show as __repr__

    # The CollimatorDatabase class has the following fields (those marked
    # with an * are set automatically and cannot be overwritten):
    #   - name
    #   - gap
    #   - jaw *
    #   - beam_size *
    #   - s_center *
    #   - angle
    #   - material
    #   - offset
    #   - tilt
    #   - stage
    #   - side
    #   - active_length
    #   - inactive_front
    #   - inactive_back
    #   - total_length *
    #   - collimator_type *
    #   - betx
    #   - bety
    #   - x
    #   - px
    #   - y
    #   - py
    #   - gamma_rel
    #   - emit

    @property
    def angle(self):
#         angles = np.array([self._colldb.angle_L.values,self._colldb.angle_R.values])
#         return pd.Series([ L if L == R else [L,R] for L, R in angles.T ], index=self._colldb.index, dtype=object)
        return self._colldb['angle_L']

    @angle.setter
    def angle(self, angle):
        self._set_property_LR('angle', angle)
        self._compute_jaws()

    @property
    def material(self):
        return self._colldb['material']

    @material.setter
    def material(self, material):
        self._set_property('material', material)

    @property
    def offset(self):
        return self._colldb['offset']

    @offset.setter
    def offset(self, offset):
        self._set_property('offset', offset, single_default_allowed=True)
        self._compute_jaws()

    @property
    def tilt(self):
        tilts = np.array([self._colldb.tilt_L.values,self._colldb.tilt_R.values])
        return pd.Series([ L if L == R else [L,R] for L, R in tilts.T ], index=self._colldb.index, dtype=object)

    @tilt.setter
    def tilt(self, tilts):
        self._set_property_LR('tilt', tilts)
        self._compute_jaws()

    @property
    def stage(self):
        return self._colldb['stage']

    @stage.setter
    def stage(self, stage):
        self._set_property('stage', stage)

    @property
    def parking(self):
        return self._colldb['parking']

    @parking.setter
    def parking(self, parking):
        self._set_property('parking', parking, single_default_allowed=True)
        self._compute_jaws()

    @property
    def active(self):
        return self._colldb['active']

    @active.setter
    def active(self, active):
        self._set_property('active', active, single_default_allowed=True)

#     @property
#     def crystal(self):
#         return self._colldb['crystal']

#     @crystal.setter
#     def crystal(self, crystal):
#         self._set_property('crystal', crystal)

#     @property
#     def bend(self):
#         return self._colldb['bend']

#     @bend.setter
#     def bend(self, bend):
#         self._set_property('bend', bend)

#     @property
#     def xdim(self):
#         return self._colldb['xdim']

#     @xdim.setter
#     def xdim(self, xdim):
#         self._set_property('xdim', xdim)

#     @property
#     def ydim(self):
#         return self._colldb['ydim']

#     @ydim.setter
#     def ydim(self, ydim):
#         self._set_property('ydim', ydim)

#     @property
#     def miscut(self):
#         return self._colldb['miscut']

#     @miscut.setter
#     def miscut(self, miscut):
#         self._set_property('miscut', miscut)

#     @property
#     def thick(self):
#         return self._colldb['thick']

#     @thick.setter
#     def thick(self, thick):
#         self._set_property('thick', thick)

    @property
    def s_center(self):
        return self._colldb['s_center']

    @property
    def collimator_type(self):
        return self._colldb['collimator_type']

    @property
    def active_length(self):
        return self._colldb['active_length']

    @active_length.setter
    def active_length(self, length):
        self._set_property('active_length', length)
        self.align_to = {}

    @property
    def inactive_front(self):
        return self._colldb['inactive_front']

    @inactive_front.setter
    def inactive_front(self, length):
        self._set_property('inactive_front', length)

    @property
    def inactive_back(self):
        return self._colldb['inactive_back']

    @inactive_back.setter
    def inactive_back(self, length):
        self._set_property('inactive_back', length)

    @property
    def total_length(self):
        return self._colldb['active_length'] +  self._colldb['inactive_front'] + self._colldb['inactive_back']

    @property
    def gap(self):
        gaps = np.array([self._colldb.gap_L.values,self._colldb.gap_R.values])
        return pd.Series([ L if L == R else [L,R] for L, R in gaps.T ], index=self._colldb.index, dtype=object)

    @gap.setter
    def gap(self, gaps):
        df = self._colldb
        correct_format = False
        # The variable gaps is a Series or a list
        if isinstance(gaps, pd.Series) or isinstance(gaps, list) or isinstance(gaps, np.ndarray):
            correct_format = True
            if len(gaps) != len(self.name):
                raise ValueError("The variable 'gaps' has a different length than the number "
                                + "of collimators in the CollimatorDatabase. Use a dictionary instead.")
            # Some of the gaps are list (e.g. two different values for both gaps): loop over gaps as dict
            if any(hasattr(gap, '__iter__') for gap in gaps):
                gaps = dict(zip(self.name, gaps))
            # All gaps are single values: use pandas-style assignment
            else:
                # mask those that have an active side for the gap under consideration
                # and have a setting less than 900; the others are set to None
                mask_L = np.logical_and(df.side.isin(['both','left']), ~(gaps >= 900))
                mask_R = np.logical_and(df.side.isin(['both','right']), ~(gaps >= 900))
                df.loc[mask_L, 'gap_L'] = gaps[mask_L]
                df.loc[~mask_L, 'gap_L'] = None
                df.loc[mask_R, 'gap_R'] = gaps[mask_R]
                df.loc[~mask_R, 'gap_R'] = None

        # The variable gaps is a dictionary
        if isinstance(gaps, dict):
            correct_format = True
            for name, gap in gaps.items():
                if name not in self.name:
                    raise ValueError(f"Collimator {name} not found in CollimatorDatabase!")
                side = df.side[name]
                if hasattr(gap, '__iter__'):
                    if isinstance(gap, str):
                        raise ValueError("The gap setting has to be a number!")
                    elif len(gap) == 2:
                        gap_L = gap[0]
                        gap_R = gap[1]
                        if side != 'both':
                            if side == 'left' and gap_R is not None:
                                print(f"Warning: collimator {name} is left-sided but a finite right gap is specified. "
                                      + "Verify that this is what you want.")
                            elif side == 'right' and gap_L is not None:
                                print(f"Warning: collimator {name} is right-sided but a finite left gap is specified. "
                                      + "Verify that this is what you want.")
                    elif len(gap) == 1:
                        gap_L = gap[0] if side in ['both','left'] else None
                        gap_R = gap[0] if side in ['both','right'] else None
                    else:
                        raise ValueError("The gap setting must have one or two values (for the left and the right jaw)!")
                else:
                    gap_L = gap if side in ['both','left'] else None
                    gap_R = gap if side in ['both','right'] else None
                gap_L = None if (gap_L is not None and gap_L >= 900) else gap_L
                gap_R = None if (gap_R is not None and gap_R >= 900) else gap_R
                df.loc[name, 'gap_L'] = gap_L
                df.loc[name, 'gap_R'] = gap_R

        if not correct_format:
            raise ValueError("Variable 'gaps' needs to be a pandas Series, dict, numpy array, or list!")

        df.gap_L = df.gap_L.astype('object', copy=False)
        df.gap_R = df.gap_R.astype('object', copy=False)
        self._compute_jaws()

    @property
    def jaw(self):
        jaws = list(np.array([
                        self._colldb.jaw_LU.values,
                        self._colldb.jaw_RU.values,
                        self._colldb.jaw_LD.values,
                        self._colldb.jaw_RD.values
                    ]).T)
        # Need special treatment if there are None's
        def flip(jaw):
            return None if jaw is None else -jaw
        for i, jaw in enumerate(jaws):
            # All 4 jaw points are the same
            if jaw[0] == flip(jaw[1]) == jaw[2] == flip(jaw[3]):
                jaws[i] = jaw[0]
            # Upstream and downstream jaws are the same
            # (all cases except angular alignment and/or tilt)
            elif jaw[0] == jaw[2] and jaw[1] == jaw[3]:
                jaws[i] = [ jaw[0], jaw[1] ]
            else:
                jaws[i] = [ [jaw[0],jaw[1]], [jaw[2],jaw[3]] ]
        return pd.Series(jaws, index=self._colldb.index, dtype=object)

    @property
    def side(self):
        return self._colldb.side

    @side.setter
    def side(self, sides):
        self._set_property('side', sides, single_default_allowed=True)
        self.gap = self.gap

    @property
    def gamma_rel(self):
        return np.sqrt(self._beta_gamma_rel**2+1)

    @gamma_rel.setter
    def gamma_rel(self, gamma_rel):
        self._beta_gamma_rel = np.sqrt(gamma_rel**2-1)
        self._compute_jaws()

    @property
    def emittance(self):
        return [self._emitx, self._emity]

    @emittance.setter
    def emittance(self, emit):
        if hasattr(emit, '__iter__'):
            if isinstance(emit, str):
                raise ValueError(f"The 'emit' setting has to be a number!")
            elif len(emit) == 2:
                self._emitx = emit[0]
                self._emity = emit[1]
            elif len(emit) == 1:
                self._emitx = emit[0]
                self._emity = emit[0]
            else:
                raise ValueError(f"The 'emit' setting must have one or two values (for emitx and emity)!")
        else:
            self._emitx = emit
            self._emity = emit
        self._compute_jaws()

    @property
    def align_to(self):
        return self._colldb.align_to

    @align_to.setter
    def align_to(self, align):
        self._set_property('align_to', align, single_default_allowed=True, limit_to=['front', 'center', 'back', 'angular'])
        if np.any(self.align_to == 'maximum'):
            raise NotImplementedError
        s_front = self.s_center - self.active_length/2
        s_center = self.s_center
        s_back = self.s_center + self.active_length/2
        mask = self.align_to == 'front'
        self._colldb.loc[mask,'s_align_front'] = s_front[mask]
        self._colldb.loc[mask,'s_align_back']  = s_front[mask]
        mask = self.align_to == 'center'
        self._colldb.loc[mask,'s_align_front'] = s_center[mask]
        self._colldb.loc[mask,'s_align_back']  = s_center[mask]
        mask = self.align_to == 'back'
        self._colldb.loc[mask,'s_align_front'] = s_back[mask]
        self._colldb.loc[mask,'s_align_back']  = s_back[mask]
        mask = self.align_to == 'angular'
        self._colldb.loc[mask,'s_align_front'] = s_front[mask]
        self._colldb.loc[mask,'s_align_back']  = s_back[mask]
        self._compute_jaws()

    # TODO: when does this need to be unset?
    @property
    def _optics_is_ready(self):
        pos = set(self._colldb.s_align_front.values) | set(self._colldb.s_align_back.values)
        return np.all([s in self._optics.index for s in pos]) and self._beta_gamma_rel is not None

    @property
    def betx(self):
        vals = np.array([
            [ self._optics.loc[s,'betx'] if s in self._optics.index else None for s in self._colldb.s_align_front.values ],
            [ self._optics.loc[s,'betx'] if s in self._optics.index else None for s in self._colldb.s_align_back.values ]
        ])
        return pd.Series([ F if F == B else [F,B] for F,B in vals.T ], index=self._colldb.index, dtype=object)

    @property
    def bety(self):
        vals = np.array([
            [ self._optics.loc[s,'bety'] if s in self._optics.index else None for s in self._colldb.s_align_front.values ],
            [ self._optics.loc[s,'bety'] if s in self._optics.index else None for s in self._colldb.s_align_back.values ]
        ])
        return pd.Series([ F if F == B else [F,B] for F,B in vals.T ], index=self._colldb.index, dtype=object)

    @property
    def alfx(self):
        vals = np.array([
            [ self._optics.loc[s,'alfx'] if s in self._optics.index else None for s in self._colldb.s_align_front.values ],
            [ self._optics.loc[s,'alfx'] if s in self._optics.index else None for s in self._colldb.s_align_back.values ]
        ])
        return pd.Series([ F if F == B else [F,B] for F,B in vals.T ], index=self._colldb.index, dtype=object)   

    @property
    def alfy(self):
        vals = np.array([
            [ self._optics.loc[s,'alfy'] if s in self._optics.index else None for s in self._colldb.s_align_front.values ],
            [ self._optics.loc[s,'alfy'] if s in self._optics.index else None for s in self._colldb.s_align_back.values ]
        ])
        return pd.Series([ F if F == B else [F,B] for F,B in vals.T ], index=self._colldb.index, dtype=object)
    
    @property
    def dx(self):
        vals = np.array([
            [ self._optics.loc[s,'dx'] if s in self._optics.index else None for s in self._colldb.s_align_front.values ],
            [ self._optics.loc[s,'dx'] if s in self._optics.index else None for s in self._colldb.s_align_back.values ]
        ])
        return pd.Series([ F if F == B else [F,B] for F,B in vals.T ], index=self._colldb.index, dtype=object)

    @property
    def dy(self):
        vals = np.array([
            [ self._optics.loc[s,'dy'] if s in self._optics.index else None for s in self._colldb.s_align_front.values ],
            [ self._optics.loc[s,'dy'] if s in self._optics.index else None for s in self._colldb.s_align_back.values ]
        ])
        return pd.Series([ F if F == B else [F,B] for F,B in vals.T ], index=self._colldb.index, dtype=object)
    
    @property
    def x(self):
        vals = np.array([
            [ self._optics.loc[s,'x'] if s in self._optics.index else None for s in self._colldb.s_align_front.values ],
            [ self._optics.loc[s,'x'] if s in self._optics.index else None for s in self._colldb.s_align_back.values ]
        ])
        return pd.Series([ F if F == B else [F,B] for F,B in vals.T ], index=self._colldb.index, dtype=object)

    @property
    def px(self):
        vals = np.array([
            [ self._optics.loc[s,'px'] if s in self._optics.index else None for s in self._colldb.s_align_front.values ],
            [ self._optics.loc[s,'px'] if s in self._optics.index else None for s in self._colldb.s_align_back.values ]
        ])
        return pd.Series([ F if F == B else [F,B] for F,B in vals.T ], index=self._colldb.index, dtype=object)

    @property
    def y(self):
        vals = np.array([
            [ self._optics.loc[s,'y'] if s in self._optics.index else None for s in self._colldb.s_align_front.values ],
            [ self._optics.loc[s,'y'] if s in self._optics.index else None for s in self._colldb.s_align_back.values ]
        ])
        return pd.Series([ F if F == B else [F,B] for F,B in vals.T ], index=self._colldb.index, dtype=object)

    @property
    def py(self):
        vals = np.array([
            [ self._optics.loc[s,'py'] if s in self._optics.index else None for s in self._colldb.s_align_front.values ],
            [ self._optics.loc[s,'py'] if s in self._optics.index else None for s in self._colldb.s_align_back.values ]
        ])
        return pd.Series([ F if F == B else [F,B] for F,B in vals.T ], index=self._colldb.index, dtype=object)

    @property
    def beam_size(self):
        if self._optics_is_ready:
            beam_size = np.array([self._beam_size_front,self._beam_size_back])
            return pd.Series([ F if F == B else [F,B] for F, B in beam_size.T ], index=self._colldb.index, dtype=object)
        else:
            return None

    @property
    def _beam_size_front(self):
        # TODO: curretnly only for angle_L
        df = self._colldb
        opt = self._optics
        betx = opt.loc[df.s_align_front,'betx'].astype(float)
        bety = opt.loc[df.s_align_front,'bety'].astype(float)
        sigmax = np.sqrt(betx*self._emitx/self._beta_gamma_rel)
        sigmay = np.sqrt(bety*self._emity/self._beta_gamma_rel)
        result = np.sqrt(
                    (sigmax*np.cos(np.float_(df.angle_L.values)*np.pi/180))**2
                    + (sigmay*np.sin(np.float_(df.angle_L.values)*np.pi/180))**2
                )
        result.index = self._colldb.index
        return result

    @property
    def _beam_size_back(self):
        # TODO: curretnly only for angle_L
        df = self._colldb
        opt = self._optics
        betx = opt.loc[df.s_align_back,'betx'].astype(float)
        bety = opt.loc[df.s_align_back,'bety'].astype(float)
        sigmax = np.sqrt(betx*self._emitx/self._beta_gamma_rel)
        sigmay = np.sqrt(bety*self._emity/self._beta_gamma_rel)
        result = np.sqrt(
                    (sigmax*np.cos(np.float_(df.angle_L.values)*np.pi/180))**2
                    + (sigmay*np.sin(np.float_(df.angle_L.values)*np.pi/180))**2
                )
        result.index = self._colldb.index
        return result

    # parking is defined with respect to closed orbit
    # TODO: tilt
    # 'upstr'  =>  'front'  en   'downstr'  =>  'back'
    def _compute_jaws(self):
        if self._optics_is_ready:
            df = self._colldb
            beam_size_front = self._beam_size_front
            beam_size_back  = self._beam_size_back
            jaw_LU = df['gap_L']*beam_size_front + self.offset
            jaw_RU = df['gap_R']*beam_size_front - self.offset
            jaw_LD = df['gap_L']*beam_size_back  + self.offset
            jaw_RD = df['gap_R']*beam_size_back  - self.offset
            df['jaw_LU'] = df['parking'] if df['gap_L'] is None else np.minimum(jaw_LU,df['parking'])
            df['jaw_RU'] = -df['parking'] if df['gap_R'] is None else -np.minimum(jaw_RU,df['parking'])
            df['jaw_LD'] = df['parking'] if df['gap_L'] is None else np.minimum(jaw_LD,df['parking'])
            df['jaw_RD'] = -df['parking'] if df['gap_R'] is None else -np.minimum(jaw_RD,df['parking'])
            # align crystals
            opt = self._optics
            df['align_angle'] = None
            cry_mask = [c is not None for c in df.crystal]
            df_cry = df[cry_mask]
            if len(df_cry) > 0:
                alfx = opt.loc[df_cry.s_align_front,'alfx'].astype(float).values
                alfy = opt.loc[df_cry.s_align_front,'alfy'].astype(float).values
                betx = opt.loc[df_cry.s_align_front,'betx'].astype(float).values
                bety = opt.loc[df_cry.s_align_front,'bety'].astype(float).values
                align_angle_x = -np.sqrt(self._emitx/self._beta_gamma_rel/betx)*alfx
                align_angle_y = -np.sqrt(self._emity/self._beta_gamma_rel/bety)*alfy
                align_angle = np.array([x if abs(ang) < 1e-6 else y
                                        for x,y,ang in zip(align_angle_x,align_angle_y,df_cry.angle_L.values)])
                df.loc[cry_mask, 'align_angle'] = align_angle*df_cry['gap_L']


    # ---------------------------------------
    # ------ Property setter functions ------
    # ---------------------------------------

    def _set_property(self, prop, vals, single_default_allowed=False, limit_to=[]):
        df = self._colldb
        if not isinstance(limit_to, (list, tuple, set)):
            limit_to = [limit_to]
        if isinstance(vals, dict):
            for name, val in vals.items():
                if name not in self.name:
                    raise ValueError(f"Collimator {name} not found in CollimatorDatabase!")
                if limit_to!=[] and val not in limit_to:
                    raise ValueError(f"Cannot set {prop} to {val}. Choose from {limit_to}!")
                df.loc[name, prop] = val
        elif isinstance(vals, pd.Series) or isinstance(vals, list) or isinstance(vals, np.ndarray):
            if len(vals) != len(self.name):
                raise ValueError(f"The variable '{prop}' has a different length than the number of "
                                + "collimators in the CollimatorDatabase. Use a dictionary instead.")
            if limit_to!=[] and np.any([val not in limit_to for val in vals]):
                raise ValueError(f"Cannot set {prop} to {vals}. Choose from {limit_to}!")
            df[prop] = vals
        else:
            if single_default_allowed:
                if limit_to!=[] and vals not in limit_to:
                    raise ValueError(f"Cannot set {prop} to {vals}. Choose from {limit_to}!")
                df[prop] = vals
            else:
                raise ValueError(f"Variable '{prop}' needs to be a pandas Series, dict, numpy array, or list!")


    def _set_property_LR(self, prop, vals):
        df = self._colldb
        correct_format = False
        # The variable vals is a Series or a list
        if isinstance(vals, pd.Series) or isinstance(vals, list) or isinstance(vals, np.ndarray):
            correct_format = True
            if len(vals) != len(self.name):
                raise ValueError(f"The variable '{prop}' has a different length than the number of "
                                + "collimators in the CollimatorDatabase. Use a dictionary instead.")
            # Some of the vals are list (e.g. two different values for both gaps): loop over vals as dict
            if any(hasattr(val, '__iter__') for val in vals):
                vals = dict(zip(self.name, vals))
            # All gaps are single values: use pandas-style assignment
            else:
                df[prop + "_L"] = vals
                df[prop + "_R"] = vals
        
        # The variable vals is a dictionary
        if isinstance(vals, dict):
            correct_format = True
            for name, val in vals.items():
                if name not in self.name:
                    raise ValueError(f"Collimator {name} not found in CollimatorDatabase!")
                if hasattr(val, '__iter__'):
                    if isinstance(val, str):
                        raise ValueError(f"The '{prop}' setting has to be a number!")
                    elif len(val) == 2:
                        val_L = val[0]
                        val_R = val[1]
                    elif len(val) == 1:
                        val_L = val[0]
                        val_R = val[0]
                    else:
                        raise ValueError(f"The '{prop}' setting must have one or two values (for the left and the right jaw)!")
                else:
                    val_L = val
                    val_R = val
                df.loc[name, prop + "_L"] = val_L
                df.loc[name, prop + "_R"] = val_R

        if not correct_format:
            raise ValueError("Variable '{prop}' needs to be a pandas Series, dict, numpy array, or list!")

        df[prop + "_L"] = df[prop + "_L"].astype('object', copy=False)
        df[prop + "_R"] = df[prop + "_R"].astype('object', copy=False)
        # Check if collimator active
        # Check if gap is list (assymetric jaws)



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xtrack as xt

from ..collimator_settings import _get_LR, _set_LR, _get_LRUD, _set_LRUD
from ..impacts import CollimatorImpacts
from ..general import _pkg_root


class InvalidXcoll(xt.BeamElement):
    _xofields = {
        'length': xo.Float64
    }

    isthick = True
    behaves_like_drift = True
    allow_track = False
    skip_in_loss_location_refinement = True
    allow_backtrack = True

    # InvalidXcoll catches unallowed cases, like backtracking through a collimator
    _extra_c_sources = [
        _pkg_root.joinpath('headers','particle_states.h'),
        _pkg_root.joinpath('headers','checks.h')
    ]

    _depends_on = [xt.RandomRutherford]  # Needed for checks

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return self.__class__(length=-self.length,
                              _context=_context, _buffer=_buffer, _offset=_offset)


class BaseBlock(xt.BeamElement):
    _xofields = {
        'length': xo.Float64
    }

    isthick = True
    allow_track = False
    behaves_like_drift = True
    skip_in_loss_location_refinement = True

    _depends_on = [InvalidXcoll]
    _internal_record_class = CollimatorImpacts

    # This is an abstract class and cannot be instantiated
    def __new__(cls, *args, **kwargs):
        if cls == BaseBlock:
            raise Exception("Abstract class `BaseBlock` cannot be instantiated!")
        instance = super().__new__(cls)
        return instance

    def enable_scattering(self):
        if hasattr(self, '_tracking'):
            self._tracking = True

    def disable_scattering(self):
        if hasattr(self, '_tracking'):
            self._tracking = False

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return InvalidXcoll(length=-self.length,
                              _context=_context, _buffer=_buffer, _offset=_offset)


class BaseCollimator(xt.BeamElement):
    _xofields = {
        'inactive_front': xo.Float64,  # Drift before jaws
        'active_length':  xo.Float64,  # Length of jaws
        'inactive_back':  xo.Float64,  # Drift after jaws
        'jaw_L':          xo.Float64,  # left jaw (distance to ref)
        'jaw_R':          xo.Float64,  # right jaw
        'ref_x':          xo.Float64,  # center of collimator reference frame
        'ref_y':          xo.Float64,
        'sin_zL':         xo.Float64,  # angle of left jaw
        'cos_zL':         xo.Float64,
        'sin_zR':         xo.Float64,  # angle of right jaw
        'cos_zR':         xo.Float64,
        'sin_yL':         xo.Float64,  # tilt of left jaw (around jaw midpoint)
        'cos_yL':         xo.Float64,
        'tan_yL':         xo.Float64,
        'sin_yR':         xo.Float64,  # tilt of right jaw (around jaw midpoint)
        'cos_yR':         xo.Float64,
        'tan_yR':         xo.Float64,
        '_side':          xo.Int8,     # is it a onesided collimator?
        'active':         xo.Int8
    }

    isthick = True
    allow_track = False
    behaves_like_drift = True
    skip_in_loss_location_refinement = True

    _skip_in_to_dict  = ['jaw_L', 'jaw_R', 'ref_x', 'ref_y',
                         'sin_yL', 'cos_yL', 'tan_yL', 'sin_yR', 'cos_yR', 'tan_yR',
                         'sin_zL', 'cos_zL', 'sin_zR', 'cos_zR', '_side']
    _store_in_to_dict = ['angle', 'tilt', 'jaw', 'reference_center', 'side']
    # Extra fields (only in Python): angle_L, angle_R, tilt_L, tilt_R, jaw_LU, jaw_LD, jaw_RU, jaw_RD

    _depends_on = [InvalidXcoll, xt.Drift, xt.XYShift, xt.SRotation, xt.YRotation]
    _internal_record_class = CollimatorImpacts

    # This is an abstract class and cannot be instantiated
    def __new__(cls, *args, **kwargs):
        if cls == BaseCollimator:
            raise Exception("Abstract class `BaseCollimator` cannot be instantiated!")
        instance = super().__new__(cls)
        return instance

    def __init__(self, **kwargs):
        if '_xobject' not in kwargs:
            # Set jaw
            if 'jaw' in kwargs:
                if 'jaw_L' in kwargs or 'jaw_R' in kwargs:
                    raise ValuError("Cannot use both 'jaw' and 'jaw_L/R'!")
                _set_LR(kwargs, 'jaw', kwargs.pop('jaw'), neg=True)
            else:
                kwargs.setdefault('jaw_L', 1)
                kwargs.setdefault('jaw_R', -1)

            # Set reference_center
            if 'reference_center' in kwargs:
                if 'ref_x' in kwargs or 'ref_y' in kwargs:
                    raise ValuError("Cannot use both 'reference_center' and 'ref_x/y'!")
                _set_LR(kwargs, 'ref', kwargs.pop('reference_center'), name_L='_x', name_R='_y')
            else:
                kwargs.setdefault('ref_x', 0)
                kwargs.setdefault('ref_y', 0)

            # Set side
            kwargs['_side'] = _side_setter(kwargs.pop('side', 'both'))

            # Set angle
            if 'angle' in kwargs:
                if 'angle_L' in kwargs or 'angle_R' in kwargs:
                    raise ValuError("Cannot use both 'angle' and 'angle_L/R'!")
                kwargs['sin_zL'], kwargs['cos_zL'], _, kwargs['sin_zR'], kwargs['cos_zR'], _ \
                            = _angle_setter(kwargs.pop('angle', 0))
            else:
                anglerad_L = kwargs.pop('angle_L', 0) / 180. * np.pi
                kwargs['sin_zL'] = np.sin(anglerad_L)
                kwargs['cos_zL'] = np.cos(anglerad_L)
                anglerad_R = kwargs.pop('angle_R', 0) / 180. * np.pi
                kwargs['sin_zR'] = np.sin(anglerad_R)
                kwargs['cos_zR'] = np.cos(anglerad_R)

            # Set tilt
            if 'tilt' in kwargs:
                if 'tilt_L' in kwargs or 'tilt_R' in kwargs:
                    raise ValuError("Cannot use both 'tilt' and 'tilt_L/R'!")
                kwargs['sin_yL'], kwargs['cos_yL'], kwargs['tan_yL'], \
                kwargs['sin_yR'], kwargs['cos_yR'], kwargs['tan_yR'] \
                                = _angle_setter(kwargs.pop('tilt', 0), rad=True)
            else:
#                 anglerad_L = kwargs.pop('tilt_L', 0) / 180. * np.pi
                anglerad_L = kwargs.pop('tilt_L', 0)
                kwargs['sin_yL'] = np.sin(anglerad_L)
                kwargs['cos_yL'] = np.cos(anglerad_L)
                kwargs['tan_yL'] = np.cos(anglerad_L)
#                 anglerad_R = kwargs.pop('tilt_R', 0) / 180. * np.pi
                anglerad_R = kwargs.pop('tilt_R', 0)
                kwargs['sin_yR'] = np.sin(anglerad_R)
                kwargs['cos_yR'] = np.cos(anglerad_R)
                kwargs['tan_yR'] = np.cos(anglerad_R)

            # Set lengths
            if 'length' in kwargs.keys():
                if 'active_length' in kwargs.keys():
                    if 'inactive_front' in kwargs.keys():
                        if 'inactive_back' in kwargs.keys():
                            raise ValueError("Too many length variables used at initialisation!")
                        kwargs['inactive_back'] = kwargs.pop('length') - kwargs['inactive_front'] - kwargs['active_length']
                    elif 'inactive_back' in kwargs.keys():
                        if 'inactive_front' in kwargs.keys():
                            raise ValueError("Too many length variables used at initialisation!")
                        kwargs['inactive_front'] = kwargs.pop('length') - kwargs['inactive_back'] - kwargs['active_length']
                    else:
                        kwargs['inactive_front'] = (kwargs['length']- kwargs['active_length'])/2
                        kwargs['inactive_back'] = (kwargs.pop('length') - kwargs['active_length'])/2
                elif 'inactive_front' in kwargs.keys():
                    if 'inactive_back' in kwargs.keys():
                        kwargs['active_length'] = kwargs.pop('length') - kwargs['inactive_back'] - kwargs['inactive_front']
                    else:
                        kwargs['active_length'] = kwargs.pop('length') - kwargs['inactive_front']
                elif 'inactive_back' in kwargs.keys():
                    kwargs['active_length'] = kwargs.pop('length') - kwargs['inactive_back']
                else:
                    kwargs['active_length'] = kwargs.pop('length')
            kwargs.setdefault('inactive_front', 0)
            kwargs.setdefault('inactive_back', 0)
            kwargs.setdefault('active_length', 0)
            kwargs.setdefault('active', True)

        super().__init__(**kwargs)


    def enable_scattering(self):
        if hasattr(self, '_tracking'):
            self._tracking = True

    def disable_scattering(self):
        if hasattr(self, '_tracking'):
            self._tracking = False

    @property
    def jaw(self):
        return _get_LR(self, 'jaw', neg=True)

    @jaw.setter
    def jaw(self, val):
        _set_LR(self, 'jaw', val, neg=True)

    @property
    def jaw_LU(self):
        return self.jaw_L - self.sin_yL*self.active_length/2

    @jaw_LU.setter   # This assumes you keep jaw_LD fixed, hence both jaw_L and the tilt change
    def jaw_LU(self, jaw_LU):
        jaw_LD = self.jaw_LD
        self.jaw_L  = (jaw_LU+jaw_LD)/2
        self.sin_yL = (jaw_LD-jaw_LU)/self.active_length
        self.cos_yL = np.sqrt(1-self.sin_yL**2)
        self.tan_yL = self.sin_yL / self.cos_yL

    @property
    def jaw_LD(self):
        return self.jaw_L + self.sin_yL*self.active_length/2

    @jaw_LD.setter   # This assumes you keep jaw_LU fixed, hence both jaw_L and the tilt change
    def jaw_LD(self, jaw_LD):
        jaw_LU = self.jaw_LU
        self.jaw_L  = (jaw_LU+jaw_LD)/2
        self.sin_yL = (jaw_LD-jaw_LU)/self.active_length
        self.cos_yL = np.sqrt(1-self.sin_yL**2)
        self.tan_yL = self.sin_yL / self.cos_yL

    @property
    def jaw_RU(self):
        return self.jaw_R - self.sin_yR*self.active_length/2

    @jaw_RU.setter   # This assumes you keep jaw_RD fixed, hence both jaw_R and the tilt change
    def jaw_RU(self, jaw_RU):
        jaw_RD = self.jaw_RD
        self.jaw_R  = (jaw_RU+jaw_RD)/2
        self.sin_yR = (jaw_RD-jaw_RU)/self.active_length
        self.cos_yR = np.sqrt(1-self.sin_yR**2)
        self.tan_yR = self.sin_yR / self.cos_yR

    @property
    def jaw_RD(self):
        return self.jaw_R + self.sin_yR*self.active_length/2

    @jaw_RD.setter   # This assumes you keep jaw_RU fixed, hence both jaw_R and the tilt change
    def jaw_RD(self, jaw_RD):
        jaw_RU = self.jaw_RU
        self.jaw_R  = (jaw_RU+jaw_RD)/2
        self.sin_yR = (jaw_RD-jaw_RU)/self.active_length
        self.cos_yR = np.sqrt(1-self.sin_yR**2)
        self.tan_yR = self.sin_yR / self.cos_yR

    @property
    def angle_L(self):
        return round(np.arctan2(self.sin_zL, self.cos_zL) * (180.0 / np.pi), 10)

    @angle_L.setter
    def angle_L(self, angle_L):
        anglerad_L = angle_L / 180. * np.pi
        self.sin_zL = np.sin(anglerad_L)
        self.cos_zL = np.cos(anglerad_L)

    @property
    def angle_R(self):
        return round(np.arctan2(self.sin_zR, self.cos_zR) * (180.0 / np.pi), 10)

    @angle_R.setter
    def angle_R(self, angle_R):
        anglerad_R = angle_R / 180. * np.pi
        self.sin_zR = np.sin(anglerad_R)
        self.cos_zR = np.cos(anglerad_R)

    @property
    def angle(self):
        return self.angle_L if self.angle_L==self.angle_R else [self.angle_L, self.angle_R]

    @angle.setter
    def angle(self, angle):
        self.sin_zL, self.cos_zL, _, self.sin_zR, self.cos_zR, _ = _angle_setter(angle)

    @property
    def reference_center(self):
        return [self.ref_x, self.ref_y]

    @reference_center.setter
    def reference_center(self, ref):
        _set_LR(self, 'ref', ref, name_L='_x', name_R='_y')

    @property
    def length(self):
        return (self.inactive_front + self.active_length + self.inactive_back)

    @length.setter
    def length(self, val):
        raise ValueError("The parameter 'length' can only be set at initialisation. " 
                       + "Use 'active_length', 'inactive_front', and/or 'inactive_back'.")

    @property
    def side(self):
        if self._side == 0:
            return 'both'
        elif self._side == 1:
            return 'left'
        elif self._side == 2:
            return 'right'

    @side.setter
    def side(self, side):
        self._side = _side_setter(side)

    # TODO: tilts are in rad! Do we want that? It's a bit inconsistent with angle which is in deg...
    # ==============================================================================================
    @property
    def tilt_L(self):
#         return round(np.arctan2(self.sin_yL, self.cos_yL) * (180.0 / np.pi), 10)
        return round(np.arctan2(self.sin_yL, self.cos_yL), 10)

    @tilt_L.setter
    def tilt_L(self, tilt_L):
#         anglerad_L = tilt_L / 180. * np.pi
        anglerad_L = tilt_L
        self.sin_yL = np.sin(anglerad_L)
        self.cos_yL = np.cos(anglerad_L)
        self.tan_yL = np.tan(anglerad_L)

    @property
    def tilt_R(self):
#         return round(np.arctan2(self.sin_yR, self.cos_yR) * (180.0 / np.pi), 10)
        return round(np.arctan2(self.sin_yR, self.cos_yR), 10)

    @tilt_R.setter
    def tilt_R(self, tilt_R):
#         anglerad_R = tilt_R / 180. * np.pi
        anglerad_R = tilt_R
        self.sin_yR = np.sin(anglerad_R)
        self.cos_yR = np.cos(anglerad_R)
        self.tan_yR = np.tan(anglerad_R)

    @property
    def tilt(self):
        return self.tilt_L if self.tilt_L==self.tilt_R else [self.tilt_L, self.tilt_R]

    @tilt.setter
    def tilt(self, tilt):
        self.sin_yL, self.cos_yL, self.tan_yL, self.sin_yR, self.cos_yR, self.tan_yR = _angle_setter(tilt, rad=True)

    def jaw_func(self, pos):
        positions = ['LU', 'RU', 'LD', 'RD']
        if not pos in positions:
            raise ValueError(f"Parameter {pos} needs to be one of {positions}!")
        point_x = getattr(self, 'ref_x')
        point_y = getattr(self, 'ref_y')
        sinz    = getattr(self, 'sin_z' + pos[0])
        cosz    = getattr(self, 'cos_z' + pos[0])
        # Shift to the jaw, whose location is given as the shortest distance:
        point_x += getattr(self, 'jaw_' + pos) * cosz
        point_y += getattr(self, 'jaw_' + pos) * sinz
        return lambda t: (point_x - t*sinz, point_y + t*cosz)


def _side_setter(val):
    if isinstance(val, str):
        if val.lower() == 'both' or val == '+-' or val == '-+':
            return 0
        elif val.lower() == 'left' or val.lower() == 'l' or val == '+':
            return 1
        elif val.lower() == 'right' or val.lower() == 'r' or val == '-':
            return 2
    raise ValueError(f"Unkown setting {val} for 'side'! Choose from "
                   + f"('left', 'L', '+'), ('right', 'R', '-'), or ('both', '+-').")

def _angle_setter(val, rad=False):
    if not hasattr(val, '__iter__'):
        val = [val]
    conversion = 1 if rad else np.pi / 180.
    if isinstance(val, str):
        raise ValueError(f"Error in setting angle: not a number!")
    elif len(val) == 2:
        anglerad_L = val[0] * conversion
        anglerad_R = val[1] * conversion
    elif len(val) == 1:
        anglerad_L = val[0] * conversion
        anglerad_R = val[0] * conversion
    else:
        raise ValueError(f"Error in setting angle: must have one or two (L, R) values!")
    return np.sin(anglerad_L), np.cos(anglerad_L), np.tan(anglerad_L), \
           np.sin(anglerad_R), np.cos(anglerad_R), np.tan(anglerad_R)





# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt

from .base import BaseBlock, BaseCollimator, InvalidXcoll
from ..scattering_routines.everest import GeneralMaterial, Material, CrystalMaterial, EverestEngine
from ..general import _pkg_root


# TODO:
#      We want these elements to behave as if 'iscollective = True' when doing twiss etc (because they would ruin the CO),
#      but as if 'iscollective = False' for normal tracking as it is natively in C...
#      Currently this is achieved with the hack '_tracking' which defaults to False after installation in the line, and is
#      only activated around the track command. Furthermore, because of 'iscollective = False' we need to specify
#      get_backtrack_element. We want it nicer..

# TODO: _per_particle_kernels should be a normal kernel (such that we don't need to pass a dummy Particles() )

class EverestBlock(BaseBlock):
    _xofields = { **BaseBlock._xofields,
        '_material':        Material,
        'rutherford_rng':   xt.RandomRutherford,
        '_tracking':        xo.Int8
    }

    isthick = True
    needs_rng = True
    allow_track = True
    behaves_like_drift = True
    skip_in_loss_location_refinement = True

    _skip_in_to_dict       = ['_material']
    _store_in_to_dict      = ['material']
    _internal_record_class = BaseBlock._internal_record_class

    _depends_on = [BaseBlock, EverestEngine]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','collimators_src','everest_block.h')
    ]

    _kernels = {
        'EverestBlock_set_material': xo.Kernel(
                c_name='EverestBlock_set_material',
                args=[xo.Arg(xo.ThisClass, name='el')]
            )
        }


    def __init__(self, **kwargs):
        if '_xobject' not in kwargs:
            mat = kwargs.pop('material', None)
            if mat is None:
                raise ValueError("Need to provide a material to the block!")
            if not isinstance(mat, Material):
                if not isinstance(mat, dict) \
                or mat['__class__'] != "Material":
                    raise ValueError("Invalid material!")
            kwargs['_material'] = mat
            kwargs.setdefault('rutherford_rng', xt.RandomRutherford())
            kwargs.setdefault('_tracking', True)
            use_prebuilt_kernels = kwargs.pop('use_prebuilt_kernels', True)
        super().__init__(**kwargs)
        if '_xobject' not in kwargs:
            self.EverestBlock_set_material(el=self)


    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, material):
        if not isinstance(material, Material):
            if not isinstance('material', dict) or material['__class__'] != "Material":
                raise ValueError("Invalid material!")
        if not xt.line._dicts_equal(self.material.to_dict(), material.to_dict()):
            self._material = material
            self.EverestBlock_set_material(el=self)

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return InvalidXcoll(length=-self.length, _context=_context,
                                 _buffer=_buffer, _offset=_offset)


class EverestCollimator(BaseCollimator):
    _xofields = { **BaseCollimator._xofields,
        '_material':        Material,
        'rutherford_rng':   xt.RandomRutherford,
        '_tracking':        xo.Int8
    }

    isthick = True
    needs_rng = True
    allow_track = True
    behaves_like_drift = True
    skip_in_loss_location_refinement = True

    _skip_in_to_dict       = [ *BaseCollimator._skip_in_to_dict, '_material' ]
    _store_in_to_dict      = [ *BaseCollimator._store_in_to_dict, 'material' ]
    _internal_record_class = BaseCollimator._internal_record_class

    _depends_on = [BaseCollimator, EverestEngine]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','collimators_src','everest_collimator.h')
    ]

    _kernels = {
        'EverestCollimator_set_material': xo.Kernel(
                c_name='EverestCollimator_set_material',
                args=[xo.Arg(xo.ThisClass, name='el')]
            )
        }


    def __init__(self, **kwargs):
        if '_xobject' not in kwargs:
            if kwargs.get('material') is None:
                raise ValueError("Need to provide a material to the collimator!")
            if not isinstance(kwargs['material'], Material):
                if not isinstance(kwargs['material'], dict) \
                or kwargs['material']['__class__'] != "Material":
                    raise ValueError("Invalid material!")
            kwargs['_material'] = kwargs.pop('material')
            kwargs.setdefault('rutherford_rng', xt.RandomRutherford())
            kwargs.setdefault('_tracking', True)
            use_prebuilt_kernels = kwargs.pop('use_prebuilt_kernels', True)
        super().__init__(**kwargs)
        if '_xobject' not in kwargs:
            self.EverestCollimator_set_material(el=self)

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, material):
        if not isinstance(material, Material):
            if not isinstance('material', dict) or material['__class__'] != "Material":
                raise ValueError("Invalid material!")
        if not xt.line._dicts_equal(self.material.to_dict(), material.to_dict()):
            self._material = material
            self.EverestCollimator_set_material(el=self)

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return InvalidXcoll(length=-self.length, _context=_context,
                                 _buffer=_buffer, _offset=_offset)



class EverestCrystal(BaseCollimator):
    _xofields = { **BaseCollimator._xofields,
        'align_angle':        xo.Float64,  #  = - sqrt(eps/beta)*alpha*nsigma
        '_bending_radius':    xo.Float64,
        '_bending_angle':     xo.Float64,
        '_critical_angle':    xo.Float64,
        'xdim':               xo.Float64,
        'ydim':               xo.Float64,
        'thick':              xo.Float64,
        'miscut':             xo.Float64,
        '_orient':            xo.Int8,
        '_material':          CrystalMaterial,
        'rutherford_rng':     xt.RandomRutherford,
        '_tracking':          xo.Int8
    }

    isthick = True
    needs_rng = True
    allow_track = True
    behaves_like_drift = True
    skip_in_loss_location_refinement = True

    _skip_in_to_dict       = [*BaseCollimator._skip_in_to_dict, '_orient', '_material', '_bending_radius',
                              '_bending_angle']
    _store_in_to_dict      = [*BaseCollimator._store_in_to_dict, 'lattice', 'material', 'bending_radius', 'bending_angle']
    _internal_record_class = BaseCollimator._internal_record_class

    _depends_on = [BaseCollimator, EverestEngine]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','collimators_src','everest_crystal.h')
    ]

    _kernels = {
        'EverestCrystal_set_material': xo.Kernel(
                c_name='EverestCrystal_set_material',
                args=[xo.Arg(xo.ThisClass, name='el')]
            )
        }


    def __init__(self, **kwargs):
        if '_xobject' not in kwargs:
            if kwargs.get('material') is None:
                raise ValueError("Need to provide a material to the collimator!")
            if not isinstance(kwargs['material'], CrystalMaterial):
                if not isinstance(kwargs['material'], dict) \
                or kwargs['material']['__class__'] != "CrystalMaterial":
                    raise ValueError("Invalid material!")
            kwargs['_material'] = kwargs.pop('material')
            bending_radius = False
            bending_angle  = False
            if 'bending_radius' in kwargs:
                if 'bending_angle' in kwargs:
                    raise ValueError("Need to choose between 'bending_radius' and 'bending_angle'!")
                bending_radius = kwargs['bending_radius']
            elif 'bending_angle' in kwargs:
                bending_angle = kwargs['bending_angle']
            kwargs['_bending_radius'] = kwargs.pop('bending_radius',0)
            kwargs['_bending_angle'] = kwargs.pop('bending_angle', 0)
            kwargs.setdefault('xdim', 0)
            kwargs.setdefault('ydim', 0)
            kwargs.setdefault('thick', 0)
            kwargs.setdefault('miscut', 0)
            kwargs['_orient'] = _lattice_setter(kwargs.pop('lattice', 'strip'))
            kwargs.setdefault('rutherford_rng', xt.RandomRutherford())
            kwargs.setdefault('_tracking', True)
            use_prebuilt_kernels = kwargs.pop('use_prebuilt_kernels', True)
        super().__init__(**kwargs)
        if '_xobject' not in kwargs:
            if bending_radius:
                self._bending_angle = np.arcsin(self.active_length/bending_radius)
            if bending_angle:
                self._bending_radius = self.active_length / np.sin(bending_angle)
            self.EverestCrystal_set_material(el=self)


    @property
    def critical_angle(self):
        return self._critical_angle if abs(self._critical_angle) > 1.e-10 else None

    @property
    def bending_radius(self):
        return self._bending_radius

    @bending_radius.setter
    def bending_radius(self, bending_radius):
        self._bending_radius = bending_radius
        self._bending_angle = np.arcsin(self.active_length/bending_radius)

    @property
    def bending_angle(self):
        return self._bending_angle

    @bending_angle.setter
    def bending_angle(self, bending_angle):
        self._bending_angle = bending_angle
        self._bending_radius = self.active_length / np.sin(bending_angle)

    @property
    def lattice(self):
        if self._orient == 1:
            return 'strip'
        elif self._orient == 2:
            return 'quasi-mosaic'
        else:
            raise ValueError(f"Illegal value {self._orient} for '_orient'!")

    @lattice.setter
    def lattice(self, lattice):
        self._orient = _lattice_setter(lattice)

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, material):
        if not isinstance(material, CrystalMaterial):
            if not isinstance(material, dict) or material['__class__'] != "CrystalMaterial":
                raise ValueError("Invalid material!")
        if not xt.line._dicts_equal(self.material.to_dict(), material.to_dict()):
            self._material = material
            self.EverestCrystal_set_material(el=self)


    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return InvalidXcoll(length=-self.length, _context=_context,
                                 _buffer=_buffer, _offset=_offset)


def _lattice_setter(lattice):
    if lattice == 'strip' or lattice == '110' or lattice == 110:
        return 1
    elif lattice == 'quasi-mosaic' or lattice == '111' or lattice == 111:
        return 2
    else:
        raise ValueError(f"Illegal value {lattice} for 'lattice'! "
                        + "Only use 'strip' (110) or 'quasi-mosaic' (111).")




# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #


tmpAuxFile = open ('colldb_lhc_run3.yaml' ,'r'  )
CDByamlflag = True

twiss_fname = args.TFile[0]
tmpOptFile = AuxFile( twiss_fname )
NREMIT = float( args.EMIn[0] )    # normalised emmittance in [m]
MOMENTUM = float( args.pc[0] )    # [GeV/c]
GEOEMIT = NREMIT / (gamma*beta)   # gamma*beta = Pc / m

# Collision IPs not done!

BeamLine = TWISS_ELEMENTs()

    # Prototype list (collimator names and positions in PARKING region)
    lbp_fname  = "prototypes.lbp" #There is a better way...but enough for now!


    # output files:
    #     1- fort.3 insertion point list
    sixt_fname = outFileName + ".fort3.list"
    #     2- insertion.txt file
    injec_fname = outFileName + ".insertion.txt"
    #     3- FLUKA input file
    out_fname  = outFileName + ".inp"
    #     4- Log file
    log_fname  = outFileName + ".log"


                element = TWISS_ELEMENT()
                COLLID += 1
                element.ID = COLLID
                element.NAME = name.upper()
                if type(settings['gap']) is str:
                    if settings['gap'] not in data['Families']:
                        error_message ( "The gap family setting '" + settings['gap'] + "' for collimator " + name + " is not defined in Families!", True )
                    element.NSIG = NullGap( data['Families'][ settings['gap'] ]['gap'] )
                else:
                    element.NSIG = NullGap( settings['gap'] )
                element.MATERIAL = settings['material']
                element.LENGTH = settings['length']
                element.ANGLE = math.radians(settings.get('angle',0))
                if float(settings.get('offset',0)) != 0:
                    print ' collimator offset ignored for the time being...'
                element.TILT1 = math.radians(settings.get('tilt_right',0))
                element.TILT2 = math.radians(settings.get('tilt_left',0))
                element.BETX = 0.0
                element.BETY = 0.0
                element.SIGX = 0.0
                element.SIGY = 0.0
                element.HALFGAP = 0.0

EXCLUDED_COLL = [ 'TCRYO' ]


# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import xobjects as xo
import xtrack as xt

from .materials import Material, CrystalMaterial
from ...impacts import CollimatorImpacts
from ...general import _pkg_root


class EverestEngine(xo.HybridClass):
    _xofields = {}

    _depends_on = [Material, CrystalMaterial, CollimatorImpacts, xt.RandomUniform, xt.RandomExponential,
                   xt.RandomNormal, xt.RandomRutherford, xt.Drift]

    _extra_c_sources = [
        _pkg_root.joinpath('scattering_routines','everest','constants.h'),
        _pkg_root.joinpath('scattering_routines','everest','everest.h'),
        _pkg_root.joinpath('scattering_routines','everest','properties.h'),
#         _pkg_root.joinpath('scattering_routines','everest','scatter_init.h'),
        _pkg_root.joinpath('scattering_routines','everest','multiple_coulomb_scattering.h'),
        _pkg_root.joinpath('scattering_routines','everest','jaw.h'),
        _pkg_root.joinpath('scattering_routines','everest','scatter.h'),
        _pkg_root.joinpath('scattering_routines','everest','crystal.h'),
        _pkg_root.joinpath('scattering_routines','everest','scatter_crystal.h')
    ]



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import xobjects as xo
import xtrack as xt

from .interaction_types import source, interactions, shortcuts
from ..general import _pkg_root

import numpy as np
import pandas as pd


class CollimatorImpacts(xt.BeamElement):
    _xofields = {
        '_index':            xt.RecordIndex,
        'at_element':        xo.Int64[:],
        'at_turn':           xo.Int64[:],
        'ds':                xo.Float64[:],
        '_inter':            xo.Int64[:],
        'parent_id':         xo.Int64[:],
        'parent_x':          xo.Float64[:],
        'parent_px':         xo.Float64[:],
        'parent_y':          xo.Float64[:],
        'parent_py':         xo.Float64[:],
        'parent_zeta':       xo.Float64[:],
        'parent_delta':      xo.Float64[:],
        'parent_energy':     xo.Float64[:],
        'parent_mass':       xo.Float64[:],
        'parent_charge':     xo.Int64[:],
        'parent_z':          xo.Int64[:],
        'parent_a':          xo.Int64[:],
        'parent_pdgid':      xo.Int64[:],
        'child_id':          xo.Int64[:],
        'child_x':           xo.Float64[:],
        'child_px':          xo.Float64[:],
        'child_y':           xo.Float64[:],
        'child_py':          xo.Float64[:],
        'child_zeta':        xo.Float64[:],
        'child_delta':       xo.Float64[:],
        'child_energy':      xo.Float64[:],
        'child_mass':        xo.Float64[:],
        'child_charge':      xo.Int64[:],
        'child_z':           xo.Int64[:],
        'child_a':           xo.Int64[:],
        'child_pdgid':       xo.Int64[:],
    }

    allow_track = False

    _extra_c_sources = [
        source,
        _pkg_root.joinpath('headers','particle_states.h'),
        _pkg_root.joinpath('impacts','impacts_src','impacts.h')
    ]

    @property
    def interaction_type(self):
        return np.array([interactions[inter] for inter in self._inter])

    def collimator_name(self, element_id):
        if not hasattr(self, '_coll_ids'):
            return element_id
        elif element_id not in self._coll_ids:
            raise ValueError(f"Element {element_id} not found in list of collimators of this impact table!\n"
                            + "Did the line change without updating the list in the impact table?")
        else:
            return self._coll_ids[element_id]

    def to_pandas(self):
        n_rows = self._index.num_recorded
        coll_header = 'collimator' if hasattr(self, '_coll_ids') else 'collimator id'
        df = pd.DataFrame({
                'turn':              self.at_turn[:n_rows],
                coll_header:         [self.collimator_name(element_id) for element_id in self.at_element[:n_rows]],
                'interaction_type':  [interactions[inter] for inter in self._inter[:n_rows]],
                'ds':                self.ds[:n_rows],
                **{
                    f'{p}_{val}': getattr(self, f'{p}_{val}')[:n_rows]
                    for p in ['parent', 'child']
                    for val in ['id', 'x', 'px', 'y', 'py', 'zeta', 'delta', 'energy', 'mass', 'charge', 'z', 'a', 'pdgid']
                }
            })
        return df

    # TODO: list of impacted collimators

    
    # TODO: does not work when multiple children
    # TODO: allow to select collimator by name
    def interactions_per_collimator(self, collimator_id=0, *, turn=None):
        mask = (self._inter > 0) & (self.at_element == collimator_id)
        if turn is not None:
            mask = mask & (self.at_turn == turn)
        df = pd.DataFrame({
                'int':  [shortcuts[inter] for inter in self._inter[mask]],
                'pid':               self.parent_id[mask]
            })
        return df.groupby('pid', sort=False)['int'].agg(list)
        
